import asyncio
from datetime import datetime, timezone
from io import BytesIO
from typing import Iterable

import pytest
from asn1crypto import cms, core, tsp
from certomancer.integrations.illusionist import Illusionist
from certomancer.registry import ArchLabel, CertLabel, KeyLabel
from freezegun import freeze_time
from pyhanko.pdf_utils.generic import pdf_name
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.pdf_utils.writer import copy_into_new_writer
from pyhanko.sign import PdfTimeStamper, fields, signers, timestamps
from pyhanko.sign.ades.api import (
    CAdESSignedAttrSpec,
    GenericCommitment,
    SignerAttrSpec,
)
from pyhanko.sign.ades.cades_asn1 import (
    CertifiedAttributeChoices,
    CommitmentTypeIndication,
    SignaturePolicyIdentifier,
    SignedAssertion,
    SignerAttributesV2,
)
from pyhanko.sign.attributes import (
    CMSAttributeProvider,
    SignedAttributeProviderSpec,
    TSTProvider,
)
from pyhanko.sign.diff_analysis import ModificationLevel
from pyhanko.sign.general import SigningError, find_cms_attribute
from pyhanko.sign.signers.pdf_signer import (
    DSSContentSettings,
    PdfTBSDocument,
    PostSignInstructions,
    SigDSSPlacementPreference,
    TimestampDSSContentSettings,
)
from pyhanko.sign.validation import (
    DocumentSecurityStore,
    PdfSignatureStatus,
    RevocationInfoValidationType,
    SignatureCoverageLevel,
    add_validation_info,
    async_validate_pdf_ltv_signature,
    async_validate_pdf_signature,
    validate_pdf_ltv_signature,
    validate_pdf_timestamp,
)
from pyhanko.sign.validation.errors import (
    SignatureValidationError,
    ValidationInfoReadingError,
)

from pyhanko_certvalidator import ValidationContext
from pyhanko_certvalidator.fetchers.requests_fetchers import (
    RequestsFetcherBackend,
)
from pyhanko_certvalidator.policy_decl import (
    CertRevTrustPolicy,
    RevocationCheckingPolicy,
    RevocationCheckingRule,
)
from pyhanko_certvalidator.registry import SimpleCertificateStore

from .samples import (
    CERTOMANCER,
    MINIMAL,
    MINIMAL_ONE_FIELD,
    MINIMAL_SLIGHTLY_BROKEN,
    MINIMAL_TWO_FIELDS,
    PDF_DATA_DIR,
    SAMPLE_GROUP_ATTR,
    TESTING_CA,
    UNRELATED_TSA,
)
from .signing_commons import (
    DUMMY_HTTP_TS,
    DUMMY_HTTP_TS_VARIANT,
    DUMMY_POLICY_ID,
    DUMMY_TS,
    DUMMY_TS2,
    FIXED_OCSP,
    FROM_CA,
    FROM_ECC_CA,
    INTERM_CERT,
    ROOT_CERT,
    SIMPLE_ECC_V_CONTEXT,
    SIMPLE_V_CONTEXT,
    TRUST_ROOTS,
    async_val_trusted,
    dummy_ocsp_vc,
    live_ac_vcs,
    live_testing_vc,
    val_trusted,
)


def ts_response_callback(request, _context):
    req = tsp.TimeStampReq.load(request.body)
    return DUMMY_TS.request_tsa_response(req=req).dump()


PADES = fields.SigSeedSubFilter.PADES


def test_pades_flag():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig1', subfilter=PADES),
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    field_name, sig_obj, sig_field = next(fields.enumerate_sig_fields(r))
    assert field_name == 'Sig1'
    assert sig_obj.get_object()['/SubFilter'] == '/ETSI.CAdES.detached'
    # the original file is a PDF 1.7 file
    assert '/ESIC' in r.root['/Extensions']
    assert r.input_version == (1, 7)


def test_pades_pdf2():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    w.ensure_output_version(version=(2, 0))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig1', subfilter=PADES),
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    field_name, sig_obj, sig_field = next(fields.enumerate_sig_fields(r))
    assert field_name == 'Sig1'
    assert sig_obj.get_object()['/SubFilter'] == '/ETSI.CAdES.detached'
    assert '/Extensions' not in r.root
    assert r.input_version == (2, 0)


@freeze_time('2020-11-01')
def test_pades_revinfo_dummydata():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=dummy_ocsp_vc(),
            subfilter=PADES,
            embed_validation_info=True,
        ),
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    field_name, sig_obj, sig_field = next(fields.enumerate_sig_fields(r))
    assert field_name == 'Sig1'
    assert sig_obj.get_object()['/SubFilter'] == '/ETSI.CAdES.detached'

    dss = DocumentSecurityStore.read_dss(handler=r)
    assert dss is not None
    assert len(dss.certs) == 4
    assert len(dss.ocsps) == 1


@freeze_time('2020-11-01')
def test_pades_revinfo_nodata():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    with pytest.raises(SigningError):
        # noinspection PyTypeChecker
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1',
                validation_context=None,
                subfilter=PADES,
                embed_validation_info=True,
            ),
            signer=FROM_CA,
        )


@freeze_time('2020-11-01')
def test_pades_revinfo_ts_dummydata():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=dummy_ocsp_vc(),
            subfilter=PADES,
            embed_validation_info=True,
        ),
        signer=FROM_CA,
        timestamper=DUMMY_TS,
    )
    r = PdfFileReader(out)
    field_name, sig_obj, sig_field = next(fields.enumerate_sig_fields(r))
    assert field_name == 'Sig1'
    assert sig_obj.get_object()['/SubFilter'] == '/ETSI.CAdES.detached'

    dss = DocumentSecurityStore.read_dss(handler=r)
    assert dss is not None
    assert len(dss.certs) == 5
    assert len(dss.ocsps) == 1


@freeze_time('2020-11-01')
def test_pades_revinfo_http_ts_dummydata(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    requests_mock.post(
        DUMMY_HTTP_TS.url,
        content=ts_response_callback,
        headers={'Content-Type': 'application/timestamp-reply'},
    )
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=dummy_ocsp_vc(),
            subfilter=PADES,
            embed_validation_info=True,
        ),
        signer=FROM_CA,
        timestamper=DUMMY_HTTP_TS,
    )
    r = PdfFileReader(out)
    field_name, sig_obj, sig_field = next(fields.enumerate_sig_fields(r))
    assert field_name == 'Sig1'
    assert sig_obj.get_object()['/SubFilter'] == '/ETSI.CAdES.detached'

    dss = DocumentSecurityStore.read_dss(handler=r)
    assert dss is not None
    assert len(dss.certs) == 5
    assert len(dss.ocsps) == 1


@freeze_time('2020-11-01')
def test_pades_revinfo_live_no_timestamp(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    vc = live_testing_vc(requests_mock)
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=vc,
            subfilter=PADES,
            embed_validation_info=True,
        ),
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    rivt_pades = RevocationInfoValidationType.PADES_LT
    with pytest.raises(ValueError):
        validate_pdf_ltv_signature(
            r.embedded_signatures[0], rivt_pades, {'trust_roots': TRUST_ROOTS}
        )


def test_pades_revinfo_live(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))

    with freeze_time('2020-11-01'):
        vc = live_testing_vc(requests_mock)
        out = signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1',
                validation_context=vc,
                subfilter=PADES,
                embed_validation_info=True,
            ),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
        )
        r = PdfFileReader(out)
        dss = DocumentSecurityStore.read_dss(handler=r)
        vc = dss.as_validation_context({})
        assert dss is not None
        assert len(dss.vri_entries) == 1
        assert len(dss.certs) == 5
        assert len(dss.ocsps) == len(vc.ocsps) == 1
        assert len(dss.crls) == len(vc.crls) == 1
        rivt_pades = RevocationInfoValidationType.PADES_LT
        status = validate_pdf_ltv_signature(
            r.embedded_signatures[0], rivt_pades, {'trust_roots': TRUST_ROOTS}
        )
        assert status.valid and status.trusted
        assert status.modification_level == ModificationLevel.LTA_UPDATES

        rivt_adobe = RevocationInfoValidationType.ADOBE_STYLE
        with pytest.raises(
            ValidationInfoReadingError, match='No revocation info'
        ):
            validate_pdf_ltv_signature(
                r.embedded_signatures[0],
                rivt_adobe,
                {'trust_roots': TRUST_ROOTS},
            )

    # test post-expiration, but before timestamp expires
    with freeze_time('2025-11-01'):
        r = PdfFileReader(out)
        status = validate_pdf_ltv_signature(
            r.embedded_signatures[0], rivt_pades, {'trust_roots': TRUST_ROOTS}
        )
        assert status.valid and status.trusted

    # test after timestamp expires: this is beyond the scope of the "basic" LTV
    #  mechanism, but failing to validate seems to be the conservative thing
    #  to do.
    with freeze_time('2040-11-01'):
        r = PdfFileReader(out)
        with pytest.raises(SignatureValidationError):
            validate_pdf_ltv_signature(
                r.embedded_signatures[0],
                rivt_pades,
                {'trust_roots': TRUST_ROOTS},
            )


@freeze_time('2020-11-01')
def test_pades_revinfo_live_update(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    vc = live_testing_vc(requests_mock)
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=vc,
            subfilter=PADES,
            embed_validation_info=True,
            use_pades_lta=True,
        ),
        signer=FROM_CA,
        timestamper=DUMMY_TS,
    )
    r = PdfFileReader(out)
    rivt_pades_lta = RevocationInfoValidationType.PADES_LTA
    # check if updates work
    out = PdfTimeStamper(DUMMY_TS).update_archival_timestamp_chain(r, vc)
    r = PdfFileReader(out)
    emb_sig = r.embedded_signatures[0]
    status = validate_pdf_ltv_signature(
        emb_sig, rivt_pades_lta, {'trust_roots': TRUST_ROOTS}
    )
    assert status.valid and status.trusted
    assert status.modification_level == ModificationLevel.LTA_UPDATES
    assert len(r.embedded_signatures) == 3
    assert len(r.embedded_regular_signatures) == 1
    assert len(r.embedded_timestamp_signatures) == 2
    assert emb_sig is r.embedded_regular_signatures[0]


@freeze_time('2020-11-01')
def test_update_no_timestamps():
    r = PdfFileReader(BytesIO(MINIMAL))
    output = PdfTimeStamper(DUMMY_TS).update_archival_timestamp_chain(
        r, dummy_ocsp_vc(), in_place=False
    )
    r = PdfFileReader(output)
    status = validate_pdf_timestamp(
        r.embedded_signatures[0],
        validation_context=ValidationContext(trust_roots=TRUST_ROOTS),
    )
    assert status.valid and status.trusted
    assert r.embedded_timestamp_signatures[0].sig_field['/TU'] == "Timestamp"


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_pades_revinfo_live_update_to_disk(requests_mock, tmp_path):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    vc = live_testing_vc(requests_mock)
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=vc,
            subfilter=PADES,
            embed_validation_info=True,
            use_pades_lta=True,
        ),
        signer=FROM_CA,
        timestamper=DUMMY_TS,
    )
    r = PdfFileReader(out)
    rivt_pades_lta = RevocationInfoValidationType.PADES_LTA
    from pathlib import Path

    out_path: Path = tmp_path / "out.pdf"
    with out_path.open('wb') as outf:
        pdf_ts = PdfTimeStamper(DUMMY_TS)
        await pdf_ts.async_update_archival_timestamp_chain(
            r, vc, in_place=False, output=outf
        )
    with out_path.open('rb') as inf:
        r = PdfFileReader(inf)
        emb_sig = r.embedded_signatures[0]
        status = await async_validate_pdf_ltv_signature(
            emb_sig, rivt_pades_lta, {'trust_roots': TRUST_ROOTS}
        )
        assert status.valid and status.trusted
        assert status.modification_level == ModificationLevel.LTA_UPDATES
        assert len(r.embedded_signatures) == 3
        assert len(r.embedded_regular_signatures) == 1
        assert len(r.embedded_timestamp_signatures) == 2
        assert emb_sig is r.embedded_regular_signatures[0]


def test_pades_revinfo_live_lta(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    _test_pades_revinfo_live_lta(w, requests_mock)


def test_pades_revinfo_live_lta_in_place(requests_mock, tmp_path):
    from pathlib import Path

    inout_file: Path = tmp_path / "test.pdf"
    inout_file.write_bytes(MINIMAL_ONE_FIELD)
    with inout_file.open('r+b') as f:
        w = IncrementalPdfFileWriter(f)
        _test_pades_revinfo_live_lta(w, requests_mock, in_place=True)


def test_pades_revinfo_live_lta_direct_flush(requests_mock, tmp_path):
    from pathlib import Path

    in_file: Path = tmp_path / "test.pdf"
    in_file.write_bytes(MINIMAL_ONE_FIELD)
    out_file: Path = tmp_path / "test-out.pdf"
    with in_file.open('rb') as inf:
        out_file.touch()
        with out_file.open('r+b') as out:
            w = IncrementalPdfFileWriter(inf)
            _test_pades_revinfo_live_lta(w, requests_mock, output=out)


def test_pades_revinfo_live_lta_direct_flush_newfile(requests_mock, tmp_path):
    # test transparent handling of non-readable/seekable output buffers
    from pathlib import Path

    in_file: Path = tmp_path / "test.pdf"
    in_file.write_bytes(MINIMAL_ONE_FIELD)
    out_file: Path = tmp_path / "test-out.pdf"
    with in_file.open('rb') as inf:
        with out_file.open('wb') as out:
            w = IncrementalPdfFileWriter(inf)
            _test_pades_revinfo_live_lta_sign(w, requests_mock, output=out)
        with out_file.open('rb') as out:
            _test_pades_revinfo_live_lta_validate(
                out, requests_mock, no_write=True
            )


def _test_pades_revinfo_live_lta_sign(w, requests_mock, **kwargs):
    with freeze_time('2020-11-01'):
        vc = live_testing_vc(requests_mock)
        out = signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1',
                validation_context=vc,
                subfilter=PADES,
                embed_validation_info=True,
                use_pades_lta=True,
            ),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
            **kwargs,
        )
    return out


def _test_pades_revinfo_live_lta_validate(
    out, requests_mock, no_write=False, has_more_sigs=False
):
    if has_more_sigs:
        expected_modlevel = ModificationLevel.FORM_FILLING
    else:
        expected_modlevel = ModificationLevel.LTA_UPDATES

    with freeze_time('2020-11-01'):
        r = PdfFileReader(out)
        dss = DocumentSecurityStore.read_dss(handler=r)
        vc = dss.as_validation_context({'trust_roots': TRUST_ROOTS})
        assert dss is not None
        if not has_more_sigs:
            assert len(dss.vri_entries) == 2
            assert len(dss.certs) == 5
            assert len(dss.ocsps) == len(vc.ocsps) == 1
            assert len(dss.crls) == len(vc.crls) == 1
        rivt_pades = RevocationInfoValidationType.PADES_LT
        status = validate_pdf_ltv_signature(
            r.embedded_signatures[0], rivt_pades, {'trust_roots': TRUST_ROOTS}
        )
        assert status.valid and status.trusted
        assert status.modification_level == expected_modlevel

        sig_obj = r.embedded_signatures[1].sig_object
        assert sig_obj.get_object()['/Type'] == pdf_name('/DocTimeStamp')

        rivt_pades_lta = RevocationInfoValidationType.PADES_LTA
        for bootstrap_vc in (None, vc):
            status = validate_pdf_ltv_signature(
                r.embedded_signatures[0],
                rivt_pades_lta,
                {'trust_roots': TRUST_ROOTS},
                bootstrap_validation_context=bootstrap_vc,
            )
            assert status.valid and status.trusted
            assert status.modification_level == expected_modlevel

    # test post-expiration, but before timestamp expires
    with freeze_time('2025-11-01'):
        r = PdfFileReader(out)
        status = validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            rivt_pades_lta,
            {'trust_roots': TRUST_ROOTS},
            bootstrap_validation_context=live_testing_vc(requests_mock),
        )
        assert status.valid and status.trusted

    # test after timestamp expires: this should fail when doing LTA testing
    with freeze_time('2035-11-01'):
        r = PdfFileReader(out)
        with pytest.raises(SignatureValidationError):
            validate_pdf_ltv_signature(
                r.embedded_signatures[0],
                rivt_pades_lta,
                {'trust_roots': TRUST_ROOTS},
                bootstrap_validation_context=live_testing_vc(requests_mock),
            )

    if no_write:
        return
    # check if updates work: use a second TSA for timestamp rollover
    with freeze_time('2028-12-01'):
        r = PdfFileReader(out)

        vc = live_testing_vc(requests_mock)
        out = PdfTimeStamper(DUMMY_TS2).update_archival_timestamp_chain(r, vc)
        r = PdfFileReader(out)
        status = validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            rivt_pades_lta,
            {'trust_roots': TRUST_ROOTS},
            bootstrap_validation_context=vc,
        )
        assert status.valid and status.trusted
        assert status.modification_level == expected_modlevel

    # the test that previously failed should now work
    with freeze_time('2035-11-01'):
        r = PdfFileReader(out)
        status = validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            rivt_pades_lta,
            {'trust_roots': TRUST_ROOTS},
            bootstrap_validation_context=live_testing_vc(requests_mock),
        )
        assert status.valid and status.trusted

    # test after timestamp expires: this should fail when doing LTA testing
    with freeze_time('2040-11-01'):
        r = PdfFileReader(out)
        with pytest.raises(SignatureValidationError):
            validate_pdf_ltv_signature(
                r.embedded_signatures[0],
                rivt_pades_lta,
                {'trust_roots': TRUST_ROOTS},
                bootstrap_validation_context=live_testing_vc(requests_mock),
            )


def _test_pades_revinfo_live_lta(w, requests_mock, **kwargs):
    out = _test_pades_revinfo_live_lta_sign(w, requests_mock, **kwargs)
    _test_pades_revinfo_live_lta_validate(out, requests_mock)


@freeze_time('2020-11-01')
def test_pades_lta_dss_indirect_arrs(requests_mock):
    testfile = PDF_DATA_DIR + '/pades-lta-dss-indirect-arrs-test.pdf'
    live_testing_vc(requests_mock)
    with open(testfile, 'rb') as f:
        r = PdfFileReader(f)
        validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            validation_type=RevocationInfoValidationType.PADES_LTA,
            # the cert embedded into this file uses a mock URL
            # that doesn't work in the current testing architecture
            validation_context_kwargs={
                'trust_roots': TRUST_ROOTS,
                'allow_fetching': False,
                'revocation_mode': 'soft-fail',
            },
        )


def test_pades_lta_sign_twice(requests_mock):
    stream = BytesIO(MINIMAL_TWO_FIELDS)
    w = IncrementalPdfFileWriter(stream)
    with freeze_time('2020-10-01'):
        vc = live_testing_vc(requests_mock)
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1',
                validation_context=vc,
                subfilter=PADES,
                embed_validation_info=True,
                use_pades_lta=True,
            ),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
            in_place=True,
        )

    w = IncrementalPdfFileWriter(stream)
    with freeze_time('2020-11-01'):
        vc = live_testing_vc(requests_mock)
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig2',
                validation_context=vc,
                subfilter=PADES,
                embed_validation_info=True,
                use_pades_lta=True,
            ),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
            in_place=True,
        )

    # test if the first sig still validates
    _test_pades_revinfo_live_lta_validate(
        stream, requests_mock, no_write=True, has_more_sigs=True
    )

    # and the second one (i.e. 3rd in the embedded_signatures list),
    # just because we can:
    with freeze_time('2025-12-01'):
        validate_pdf_ltv_signature(
            PdfFileReader(stream).embedded_signatures[2],
            validation_type=RevocationInfoValidationType.PADES_LTA,
            validation_context_kwargs={'trust_roots': TRUST_ROOTS},
            bootstrap_validation_context=live_testing_vc(requests_mock),
        )


def test_pades_lta_sign_twice_post_expiry(requests_mock):
    stream = BytesIO(MINIMAL_TWO_FIELDS)
    w = IncrementalPdfFileWriter(stream)
    with freeze_time('2020-10-01'):
        vc = live_testing_vc(requests_mock)
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1',
                validation_context=vc,
                subfilter=PADES,
                embed_validation_info=True,
                use_pades_lta=True,
            ),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
            in_place=True,
        )

    w = IncrementalPdfFileWriter(stream)
    with freeze_time('2020-10-10'):
        # intentionally load a VC in which the original TS does
        # not validate
        vc = SIMPLE_ECC_V_CONTEXT()
        with pytest.raises(SigningError, match=".*most recent timestamp.*"):
            signers.sign_pdf(
                w,
                signers.PdfSignatureMetadata(
                    field_name='Sig2',
                    validation_context=vc,
                    subfilter=PADES,
                    embed_validation_info=True,
                    use_pades_lta=True,
                ),
                signer=FROM_ECC_CA,
                timestamper=DUMMY_TS,
                in_place=True,
            )


@freeze_time('2020-11-01')
def test_standalone_document_timestamp(requests_mock):
    pdf_ts = signers.PdfTimeStamper(timestamper=DUMMY_TS)
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    vc = live_testing_vc(requests_mock)
    out = pdf_ts.timestamp_pdf(w, md_algorithm='sha256', validation_context=vc)
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    with pytest.raises(SignatureValidationError, match='.*must be /Sig.*'):
        val_trusted(s, vc=vc)

    status = validate_pdf_timestamp(embedded_sig=s, validation_context=vc)
    assert status.valid and status.trusted
    assert status.coverage == SignatureCoverageLevel.ENTIRE_REVISION
    assert status.modification_level == ModificationLevel.LTA_UPDATES

    # tamper with the file
    out.seek(0x9D)
    out.write(b'4')
    out.seek(0)
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    tampered = validate_pdf_timestamp(embedded_sig=s, validation_context=vc)
    assert not tampered.intact and tampered.valid


@pytest.mark.parametrize('with_vri', [True, False])
def test_add_revinfo_later(requests_mock, with_vri):
    buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(buf)

    # create signature without revocation info
    with freeze_time('2020-11-01'):
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig1'),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
            in_place=True,
        )

    # fast forward 1 month
    with freeze_time('2020-12-01'):
        vc = live_testing_vc(requests_mock)
        r = PdfFileReader(buf)
        emb_sig = r.embedded_signatures[0]
        add_validation_info(emb_sig, vc, in_place=True, add_vri_entry=with_vri)

        r = PdfFileReader(buf)
        emb_sig = r.embedded_signatures[0]

        # without retroactive revinfo, the validation should fail
        status = validate_pdf_ltv_signature(
            emb_sig,
            RevocationInfoValidationType.PADES_LT,
            {'trust_roots': TRUST_ROOTS},
        )
        assert status.valid and not status.trusted

        # with retroactive revinfo, it should be OK
        status = validate_pdf_ltv_signature(
            emb_sig,
            RevocationInfoValidationType.PADES_LT,
            {'trust_roots': TRUST_ROOTS, 'retroactive_revinfo': True},
        )
        assert status.valid and status.trusted
        assert status.modification_level == ModificationLevel.LTA_UPDATES


@pytest.mark.parametrize('with_vri', [True, False])
def test_fix_incomplete_revinfo_later(requests_mock, with_vri):
    buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(buf)

    # create signature with incomplete revinfo
    with freeze_time('2020-11-01'):
        # trust the intermediate cert and the TSA cert absolutely
        incomplete_vc = ValidationContext(
            trust_roots=[INTERM_CERT, DUMMY_TS.tsa_cert],
            allow_fetching=True,
            other_certs=[],
        )
        Illusionist(TESTING_CA).register(requests_mock)
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1',
                validation_context=incomplete_vc,
                embed_validation_info=True,
                subfilter=PADES,
            ),
            signer=FROM_CA,
            timestamper=DUMMY_TS,
            in_place=True,
        )

    # fast forward 1 month
    with freeze_time('2020-12-01'):
        vc = live_testing_vc(requests_mock)
        r = PdfFileReader(buf)
        emb_sig = r.embedded_signatures[0]
        add_validation_info(emb_sig, vc, in_place=True, add_vri_entry=with_vri)

        r = PdfFileReader(buf)
        emb_sig = r.embedded_signatures[0]

        status = validate_pdf_ltv_signature(
            emb_sig,
            RevocationInfoValidationType.PADES_LT,
            {'trust_roots': TRUST_ROOTS, 'retroactive_revinfo': True},
        )
        assert status.valid and status.trusted
        assert status.modification_level == ModificationLevel.LTA_UPDATES


def test_add_revinfo_and_timestamp(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))

    # create signature without revocation info
    with freeze_time('2020-11-01'):
        out = signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig1'),
            signer=FROM_CA,
            in_place=True,
        )

    # fast forward 1 month
    with freeze_time('2020-12-01'):
        vc = live_testing_vc(requests_mock)
        r = PdfFileReader(out)
        emb_sig = r.embedded_signatures[0]
        out = add_validation_info(emb_sig, vc)

        timestamper = signers.PdfTimeStamper(timestamper=DUMMY_TS)
        timestamper.timestamp_pdf(
            IncrementalPdfFileWriter(out), 'sha256', vc, in_place=True
        )

        r = PdfFileReader(out)
        emb_sig = r.embedded_signatures[0]

        # This should suffice for PAdES-LT, even without retroactive_revinfo
        # (since the new timestamp is now effectively the only trusted record
        #  of the signing time anyway)
        status = validate_pdf_ltv_signature(
            emb_sig,
            RevocationInfoValidationType.PADES_LT,
            {'trust_roots': TRUST_ROOTS},
        )
        assert status.valid and status.trusted
        assert status.signer_reported_dt == datetime.now(tz=timezone.utc)

        # ... but PAdES-LTA should fail
        with pytest.raises(
            SignatureValidationError, match='.*requires separate timestamps.*'
        ):
            validate_pdf_ltv_signature(
                emb_sig,
                RevocationInfoValidationType.PADES_LTA,
                {'trust_roots': TRUST_ROOTS},
            )


def test_add_revinfo_and_lta_timestamp(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))

    # create signature without revocation info
    with freeze_time('2020-11-01'):
        out = signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig1'),
            signer=FROM_CA,
            in_place=True,
        )

    # fast forward 1 month
    with freeze_time('2020-12-01'):
        vc = live_testing_vc(requests_mock)
        r = PdfFileReader(out)
        emb_sig = r.embedded_signatures[0]
        out = add_validation_info(emb_sig, vc)

        timestamper = signers.PdfTimeStamper(timestamper=DUMMY_TS)
        timestamper.timestamp_pdf(
            IncrementalPdfFileWriter(out), 'sha256', vc, in_place=True
        )
        timestamper.update_archival_timestamp_chain(PdfFileReader(out), vc)

        r = PdfFileReader(out)
        emb_sig = r.embedded_signatures[0]

        status = validate_pdf_ltv_signature(
            emb_sig,
            RevocationInfoValidationType.PADES_LTA,
            {'trust_roots': TRUST_ROOTS},
        )
        assert status.valid and status.trusted
        assert status.signer_reported_dt == datetime.now(tz=timezone.utc)

    # test post-expiration, but before timestamp expires
    with freeze_time('2025-11-01'):
        r = PdfFileReader(out)
        status = validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            RevocationInfoValidationType.PADES_LTA,
            {'trust_roots': TRUST_ROOTS},
            bootstrap_validation_context=live_testing_vc(requests_mock),
        )
        assert status.valid and status.trusted


@freeze_time('2020-11-01')
def test_sign_with_commitment():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1',
        subfilter=fields.SigSeedSubFilter.PADES,
        cades_signed_attr_spec=CAdESSignedAttrSpec(
            commitment_type=GenericCommitment.PROOF_OF_ORIGIN.asn1
        ),
    )
    out = signers.sign_pdf(w, meta, signer=FROM_CA)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb)

    indic = find_cms_attribute(
        emb.signer_info['signed_attrs'], 'commitment_type'
    )[0]
    assert indic['commitment_type_id'].native == 'proof_of_origin'


@freeze_time('2020-11-01')
def test_sign_with_content_sig():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1',
        subfilter=fields.SigSeedSubFilter.PADES,
        cades_signed_attr_spec=CAdESSignedAttrSpec(timestamp_content=True),
    )
    out = signers.sign_pdf(w, meta, signer=FROM_CA, timestamper=DUMMY_TS)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    status = val_trusted(emb)
    assert isinstance(status, PdfSignatureStatus)
    assert status.content_timestamp_validity.intact
    assert status.content_timestamp_validity.valid
    assert status.content_timestamp_validity.trusted

    content_ts = find_cms_attribute(
        emb.signer_info['signed_attrs'], 'content_time_stamp'
    )[0]
    eci = content_ts['content']['encap_content_info']
    assert eci['content_type'].native == 'tst_info'


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_with_wrong_content_sig():
    class Spec(SignedAttributeProviderSpec):
        def signed_attr_providers(
            self, data_digest: bytes, digest_algorithm: str
        ) -> Iterable[CMSAttributeProvider]:
            yield TSTProvider(
                digest_algorithm='sha256',
                data_to_ts=b'\xde\xad\xbe\xef',
                timestamper=DUMMY_TS,
                attr_type='content_time_stamp',
            )

    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1',
        subfilter=fields.SigSeedSubFilter.PADES,
    )
    signer = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
    )
    signer.signed_attr_prov_spec = Spec()
    out = await signers.async_sign_pdf(
        w, meta, signer=signer, timestamper=DUMMY_TS
    )

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    status = await async_validate_pdf_signature(
        embedded_sig=emb, signer_validation_context=SIMPLE_V_CONTEXT()
    )
    assert not status.bottom_line
    assert status.valid
    assert status.intact
    assert status.trusted
    assert not status.content_timestamp_validity.intact
    assert status.content_timestamp_validity.valid
    assert not status.content_timestamp_validity.trusted
    assert status.timestamp_validity.intact
    assert status.timestamp_validity.valid
    assert status.timestamp_validity.trusted


@freeze_time('2020-11-01')
@pytest.mark.parametrize('different_tsa', [True, False])
def test_interrupted_pades_lta_signature(requests_mock, different_tsa):
    # simulate a PAdES-LTA workflow with remote signing
    # (our hypothetical remote signer just signs digests, not full CMS objects)
    requests_mock.post(
        DUMMY_HTTP_TS.url,
        content=ts_response_callback,
        headers={'Content-Type': 'application/timestamp-reply'},
    )

    def instantiate_external_signer(sig_value: bytes):
        return signers.ExternalSigner(
            signing_cert=TESTING_CA.get_cert(CertLabel('signer1')),
            signature_value=sig_value,
            cert_registry=SimpleCertificateStore.from_certs(
                [ROOT_CERT, INTERM_CERT]
            ),
        )

    async def prep_doc():
        # 2048-bit RSA sig is 256 bytes long -> placeholder
        ext_signer = instantiate_external_signer(sig_value=bytes(256))
        vc = live_testing_vc(requests_mock)
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
        pdf_signer = signers.PdfSigner(
            signers.PdfSignatureMetadata(
                field_name='SigNew',
                embed_validation_info=True,
                use_pades_lta=True,
                subfilter=PADES,
                validation_context=vc,
                md_algorithm='sha256',
                dss_settings=DSSContentSettings(include_vri=False),
            ),
            signer=ext_signer,
            timestamper=DUMMY_HTTP_TS,
        )
        # This function may perform async network I/O
        (
            prep_digest,
            tbs_document,
            output,
        ) = await pdf_signer.async_digest_doc_for_signing(w)
        psi = tbs_document.post_sign_instructions

        # prepare signed attributes
        # Note: signed attribute construction may involve
        # expensive I/O (e.g. when content timestamp tokens
        # need to be obtained). Hence why the API is asynchronous.
        signed_attrs = await ext_signer.signed_attrs(
            prep_digest.document_digest, 'sha256', use_pades=True
        )
        return prep_digest, signed_attrs, psi, output

    async def sim_sign_remote(data_tbs: bytes):
        # pretend that everything below happens on a remote server
        # (we declare the function as async to add to the illusion)
        from cryptography.hazmat.primitives import hashes, serialization
        from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15

        priv_key = serialization.load_der_private_key(
            TESTING_CA.key_set.get_private_key(KeyLabel('signer1')).dump(),
            password=None,
        )
        padding = PKCS1v15()
        hash_algo = hashes.SHA256()
        return priv_key.sign(data_tbs, padding, hash_algo)

    async def proceed_with_signing(
        prep_digest, signed_attrs, sig_value, psi, output_handle
    ):
        # use ExternalSigner to format the CMS given the signed value
        # we obtained from the remote signing service
        ext_signer = instantiate_external_signer(sig_value)

        # again, even though we already produced the signature, we need to
        # perform an async call here, because setting up the unsigned attributes
        # potentially requires async I/O in the general case
        sig_cms = await ext_signer.async_sign_prescribed_attributes(
            'sha256', signed_attrs=signed_attrs, timestamper=DUMMY_HTTP_TS
        )

        # fresh VC
        validation_context = live_testing_vc(
            requests_mock, with_extra_tsa=different_tsa
        )
        await PdfTBSDocument.async_finish_signing(
            output_handle,
            prepared_digest=prep_digest,
            signature_cms=sig_cms,
            post_sign_instr=psi,
            validation_context=validation_context,
        )

    async def full_procedure():
        # prepare document and signed attributes to sign
        prep_digest, signed_attrs, psi, output = await prep_doc()

        # copy the output to a new buffer, just to drive home the point that
        # we no longer care about the original output stream
        new_output = BytesIO()
        assert isinstance(output, BytesIO)
        buf = output.getbuffer()
        new_output.write(buf)
        buf.release()

        # contact our 'remote' signing service
        sig_value = await sim_sign_remote(signed_attrs.dump())
        # let's pretend that we no longer have access to the non-serialisable
        # parts of tbs_document and psi (e.g. because this part happens on a
        # different machine)
        if different_tsa:
            # use another TSA for the document timestamp
            docts_tsa = DUMMY_HTTP_TS_VARIANT
        else:
            docts_tsa = DUMMY_HTTP_TS

        new_psi = PostSignInstructions(
            validation_info=psi.validation_info,
            timestamp_md_algorithm=psi.timestamp_md_algorithm,
            timestamper=docts_tsa,
            timestamp_field_name=psi.timestamp_field_name,
            dss_settings=psi.dss_settings,
        )
        # finish signing
        await proceed_with_signing(
            prep_digest, signed_attrs, sig_value, new_psi, new_output
        )
        return new_output

    r = PdfFileReader(asyncio.run(full_procedure()))
    # check cardinality of DSS content
    dss = DocumentSecurityStore.read_dss(handler=r)
    assert dss is not None
    if different_tsa:
        assert len(dss.certs) == 7
        assert len(dss.ocsps) == 1
        assert len(dss.crls) == 2
        trust_roots = TRUST_ROOTS + [UNRELATED_TSA.get_cert(CertLabel('root'))]
        # extra update was needed to append revinfo for second TSA
        assert r.xrefs.total_revisions == 4
    else:
        assert len(dss.certs) == 5
        assert len(dss.ocsps) == 1
        assert len(dss.crls) == 1
        assert r.xrefs.total_revisions == 3
        trust_roots = TRUST_ROOTS

    emb_sig = r.embedded_signatures[0]
    # perform LTA validation
    status = validate_pdf_ltv_signature(
        emb_sig,
        RevocationInfoValidationType.PADES_LTA,
        {'trust_roots': trust_roots},
    )
    assert status.valid and status.trusted
    assert status.modification_level == ModificationLevel.LTA_UPDATES
    assert len(r.embedded_signatures) == 2
    assert len(r.embedded_regular_signatures) == 1
    assert len(r.embedded_timestamp_signatures) == 1
    assert emb_sig is r.embedded_regular_signatures[0]


def test_dss_setting_validation():
    bad_ts_sett = TimestampDSSContentSettings(
        include_vri=True, update_before_ts=True
    )
    with pytest.raises(SigningError):
        bad_ts_sett.assert_viable()

    with pytest.raises(SigningError):
        DSSContentSettings(
            include_vri=True,
            placement=SigDSSPlacementPreference.TOGETHER_WITH_SIGNATURE,
        ).assert_viable()

    DSSContentSettings(
        include_vri=True,
        placement=SigDSSPlacementPreference.TOGETHER_WITH_NEXT_TS,
    ).assert_viable()
    DSSContentSettings(
        include_vri=False,
        placement=SigDSSPlacementPreference.TOGETHER_WITH_SIGNATURE,
    ).assert_viable()
    with pytest.raises(SigningError):
        DSSContentSettings(
            include_vri=False,
            placement=SigDSSPlacementPreference.TOGETHER_WITH_SIGNATURE,
            next_ts_settings=bad_ts_sett,
        ).assert_viable()


@freeze_time('2020-11-01')
def test_pades_one_revision(requests_mock):
    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL_ONE_FIELD)))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            dss_settings=DSSContentSettings(
                include_vri=False,
                placement=SigDSSPlacementPreference.TOGETHER_WITH_SIGNATURE,
            ),
            validation_context=live_testing_vc(requests_mock),
            embed_validation_info=True,
        ),
        timestamper=DUMMY_TS,
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    assert r.total_revisions == 1
    validate_pdf_ltv_signature(
        r.embedded_signatures[0],
        validation_type=RevocationInfoValidationType.PADES_LT,
        validation_context_kwargs={
            'trust_roots': TRUST_ROOTS,
            'allow_fetching': False,
            'revocation_mode': 'soft-fail',
        },
    )


NOOP_POLICY = CertRevTrustPolicy(
    revocation_checking_policy=RevocationCheckingPolicy(
        ee_certificate_rule=RevocationCheckingRule.NO_CHECK,
        intermediate_ca_cert_rule=RevocationCheckingRule.NO_CHECK,
    )
)


def _lazy_pades_signature(requests_mock):
    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL_ONE_FIELD)))
    # set up a signer that doesn't embed anything
    #  (but still goes through the motions)
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            validation_context=live_testing_vc(
                requests_mock, revinfo_policy=NOOP_POLICY
            ),
            embed_validation_info=True,
            use_pades_lta=True,
        ),
        timestamper=DUMMY_TS,
        signer=FROM_CA,
    )
    return out


@freeze_time('2020-11-01')
def test_pades_ltv_legacy_policy_sufficient(requests_mock):
    out = _lazy_pades_signature(requests_mock)
    r = PdfFileReader(out)
    # soft fail should not apply to the internal timestamp, so we expect
    # validation to fail
    with pytest.raises(SignatureValidationError, match='time of signing'):
        validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            validation_type=RevocationInfoValidationType.PADES_LTA,
            validation_context_kwargs={
                'trust_roots': TRUST_ROOTS,
                'allow_fetching': False,
                'revocation_mode': 'hard-fail',
            },
            # allow bootstrapping with soft-fail
            bootstrap_validation_context=ValidationContext(
                trust_roots=TRUST_ROOTS,
                allow_fetching=False,
                revocation_mode='soft-fail',
            ),
        )


@freeze_time('2020-11-01')
def test_pades_ltv_upgrade_soft_fail(requests_mock):
    out = _lazy_pades_signature(requests_mock)
    r = PdfFileReader(out)
    # soft fail should not apply to the internal timestamp, so we expect
    # validation to fail
    with pytest.raises(SignatureValidationError, match='time of signing'):
        validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            validation_type=RevocationInfoValidationType.PADES_LTA,
            validation_context_kwargs={
                'trust_roots': TRUST_ROOTS,
                'allow_fetching': False,
                'revocation_mode': 'soft-fail',
            },
        )


@freeze_time('2020-11-01')
def test_pades_ltv_upgrade_lax_policy(requests_mock):
    out = _lazy_pades_signature(requests_mock)
    r = PdfFileReader(out)
    # as in the soft_fail case, we expect this to fail
    with pytest.raises(SignatureValidationError, match='time of signing'):
        validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            validation_type=RevocationInfoValidationType.PADES_LTA,
            validation_context_kwargs={
                'trust_roots': TRUST_ROOTS,
                'allow_fetching': False,
                'revinfo_policy': NOOP_POLICY,
            },
        )


@freeze_time('2020-11-01')
@pytest.mark.parametrize(
    'dss_settings',
    [
        DSSContentSettings(
            placement=SigDSSPlacementPreference.SEPARATE_REVISION,
        ),
        DSSContentSettings(
            placement=SigDSSPlacementPreference.TOGETHER_WITH_NEXT_TS,
        ),
        DSSContentSettings(),
    ],
)
def test_pades_two_revisions(requests_mock, dss_settings):
    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL_ONE_FIELD)))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            dss_settings=dss_settings,
            validation_context=live_testing_vc(requests_mock),
            embed_validation_info=True,
        ),
        timestamper=DUMMY_TS,
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    assert r.total_revisions == 2
    validate_pdf_ltv_signature(
        r.embedded_signatures[0],
        validation_type=RevocationInfoValidationType.PADES_LT,
        validation_context_kwargs={
            'trust_roots': TRUST_ROOTS,
            'allow_fetching': False,
            'revocation_mode': 'soft-fail',
        },
    )


@freeze_time('2020-11-01')
@pytest.mark.parametrize(
    'dss_settings',
    [
        DSSContentSettings(
            placement=SigDSSPlacementPreference.TOGETHER_WITH_NEXT_TS,
            next_ts_settings=TimestampDSSContentSettings(
                update_before_ts=True, include_vri=False
            ),
        ),
        DSSContentSettings(
            placement=SigDSSPlacementPreference.TOGETHER_WITH_SIGNATURE,
            include_vri=False,
            next_ts_settings=TimestampDSSContentSettings(
                update_before_ts=True, include_vri=False
            ),
        ),
        DSSContentSettings(
            placement=SigDSSPlacementPreference.TOGETHER_WITH_SIGNATURE,
            include_vri=False,
        ),
        DSSContentSettings(
            placement=SigDSSPlacementPreference.TOGETHER_WITH_NEXT_TS,
            include_vri=False,
        ),
    ],
)
def test_pades_lta_two_revisions(requests_mock, dss_settings):
    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL_ONE_FIELD)))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            dss_settings=dss_settings,
            validation_context=live_testing_vc(requests_mock),
            embed_validation_info=True,
            use_pades_lta=True,
        ),
        timestamper=DUMMY_TS,
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    assert r.total_revisions == 2
    validate_pdf_ltv_signature(
        r.embedded_signatures[0],
        validation_type=RevocationInfoValidationType.PADES_LTA,
        validation_context_kwargs={
            'trust_roots': TRUST_ROOTS,
            'allow_fetching': False,
            'revocation_mode': 'soft-fail',
        },
    )


@freeze_time('2020-11-01')
def test_pades_lta_noskip(requests_mock):
    dss_settings = DSSContentSettings(
        placement=SigDSSPlacementPreference.SEPARATE_REVISION,
        include_vri=False,
        skip_if_unneeded=False,
    )
    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL_ONE_FIELD)))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            dss_settings=dss_settings,
            validation_context=live_testing_vc(requests_mock),
            embed_validation_info=True,
            use_pades_lta=True,
        ),
        timestamper=DUMMY_TS,
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    assert r.total_revisions == 4
    validate_pdf_ltv_signature(
        r.embedded_signatures[0],
        validation_type=RevocationInfoValidationType.PADES_LTA,
        validation_context_kwargs={
            'trust_roots': TRUST_ROOTS,
            'allow_fetching': False,
            'revocation_mode': 'soft-fail',
        },
    )


@freeze_time('2020-11-01')
def test_pades_post_ts_autosuppress(requests_mock):
    # test if the post-timestamp DSS update is automatically suppressed
    # if we sign with the exact same TSA for both the sig & the document TS
    # (keeping the VC constant as well)
    dss_settings = DSSContentSettings(
        placement=SigDSSPlacementPreference.TOGETHER_WITH_NEXT_TS,
        include_vri=False,
    )
    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL_ONE_FIELD)))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            dss_settings=dss_settings,
            validation_context=live_testing_vc(requests_mock),
            embed_validation_info=True,
            use_pades_lta=True,
        ),
        timestamper=DUMMY_TS,
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    assert r.total_revisions == 2
    # assert that the DSS was updated in the timestamped revision
    dss_ref = r.root.get_value_as_reference('/DSS')
    changed_in = r.xrefs.get_last_change(dss_ref)
    assert changed_in == 1

    validate_pdf_ltv_signature(
        r.embedded_signatures[0],
        validation_type=RevocationInfoValidationType.PADES_LTA,
        validation_context_kwargs={
            'trust_roots': TRUST_ROOTS,
            'allow_fetching': False,
            'revocation_mode': 'soft-fail',
        },
    )


@freeze_time('2020-11-01')
def test_pades_max_autosuppress(requests_mock):
    # test if all timestamp-related DSS updates are suppressed
    # if we sign with the exact same TSA for both the sig & the document TS
    # (keeping the VC constant as well) and all relevant DSS updates were
    # performed pre-signing
    dss_settings = DSSContentSettings(
        placement=SigDSSPlacementPreference.TOGETHER_WITH_SIGNATURE,
        include_vri=False,
    )
    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL_ONE_FIELD)))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            dss_settings=dss_settings,
            validation_context=live_testing_vc(requests_mock),
            embed_validation_info=True,
            use_pades_lta=True,
        ),
        timestamper=DUMMY_TS,
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    assert r.total_revisions == 2
    # assert that the DSS was not updated in the timestamped revision
    dss_ref = r.root.get_value_as_reference('/DSS')
    changed_in = r.xrefs.get_last_change(dss_ref)
    assert changed_in == 0

    validate_pdf_ltv_signature(
        r.embedded_signatures[0],
        validation_type=RevocationInfoValidationType.PADES_LTA,
        validation_context_kwargs={
            'trust_roots': TRUST_ROOTS,
            'allow_fetching': False,
            'revocation_mode': 'soft-fail',
        },
    )


@freeze_time('2020-11-01')
def test_pades_independent_tsa(requests_mock):
    # test signing/validation behaviour with an independent TSA

    w = copy_into_new_writer(PdfFileReader(BytesIO(MINIMAL_ONE_FIELD)))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            validation_context=live_testing_vc(
                requests_mock, with_extra_tsa=True
            ),
            embed_validation_info=True,
        ),
        signer=FROM_CA,
        timestamper=DUMMY_HTTP_TS_VARIANT,
    )
    r = PdfFileReader(out)

    # DSS cardinalities
    dss_dict = r.root['/DSS']
    assert len(dss_dict['/CRLs']) == 2
    assert len(dss_dict['/OCSPs']) == 1
    assert len(dss_dict['/Certs']) == 6

    assert r.xrefs.total_revisions == 2

    trust_roots = TRUST_ROOTS + [UNRELATED_TSA.get_cert(CertLabel('root'))]
    validate_pdf_ltv_signature(
        r.embedded_signatures[0],
        validation_type=RevocationInfoValidationType.PADES_LT,
        validation_context_kwargs={
            'trust_roots': trust_roots,
            'allow_fetching': False,
            'revocation_mode': 'soft-fail',
        },
    )


@freeze_time('2020-11-01')
def test_sign_with_policy():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1',
        subfilter=fields.SigSeedSubFilter.PADES,
        cades_signed_attr_spec=CAdESSignedAttrSpec(
            signature_policy_identifier=SignaturePolicyIdentifier(
                {'signature_policy_id': DUMMY_POLICY_ID}
            )
        ),
    )
    out = signers.sign_pdf(w, meta, signer=FROM_CA)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb)

    sp_id = find_cms_attribute(
        emb.signer_info['signed_attrs'], 'signature_policy_identifier'
    )[0]
    assert sp_id.chosen['sig_policy_id'].native == '2.999'


@freeze_time('2020-11-01')
def test_pades_revinfo_live_nofullchain():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=dummy_ocsp_vc(),
            subfilter=PADES,
            embed_validation_info=True,
        ),
        signer=FROM_CA,
        timestamper=DUMMY_TS,
    )
    r = PdfFileReader(out)
    rivt_pades = RevocationInfoValidationType.PADES_LT

    # with the same dumb settings, the timestamp doesn't validate at all,
    # which causes LTV validation to fail to bootstrap
    with pytest.raises(SignatureValidationError):
        validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            rivt_pades,
            {
                'trust_roots': TRUST_ROOTS,
                'ocsps': [FIXED_OCSP],
                'allow_fetching': False,
            },
        )

    # now set up live testing
    from requests_mock import Mocker

    with Mocker() as m:
        live_testing_vc(m)
        status = validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            rivt_pades,
            {'trust_roots': TRUST_ROOTS, 'allow_fetching': True},
        )
        # .. which should still fail because the chain of trust is broken, but
        # at least the timestamp should initially validate
        assert status.valid and not status.trusted, status.summary()


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_pades_lta_no_embed_root(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    cr = SimpleCertificateStore()
    cr.register_multiple(FROM_CA.cert_registry)
    vc = live_testing_vc(requests_mock)
    no_embed_root_signer = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=cr,
        embed_roots=False,
    )
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            embed_validation_info=True,
            use_pades_lta=True,
            validation_context=vc,
        ),
        signer=no_embed_root_signer,
        timestamper=DUMMY_HTTP_TS,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    assert '/AP' not in s.sig_field
    # signer, intermediate, but not TSA (that one is supposed to be in the
    # TST) and of course no root
    assert len(s.signed_data['certificates']) == 2
    # signer, intermediate, TSA and OCSP responder
    assert len(r.root['/DSS']['/Certs']) == 4
    await async_validate_pdf_ltv_signature(
        r.embedded_signatures[0],
        validation_type=RevocationInfoValidationType.PADES_LTA,
        validation_context_kwargs={'trust_roots': TRUST_ROOTS},
    )


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_pades_live_ac_presign_validation(requests_mock):
    # integration test for heavy-duty autofetching logic with ACs
    # NOTE: certificate autofetching is not tested due to lack of availability
    # in Illusionist (at the time of writing)

    pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))
    authorities = [
        pki_arch.get_cert('root'),
        pki_arch.get_cert('interm'),
        pki_arch.get_cert('root-aa'),
        pki_arch.get_cert('interm-aa'),
        pki_arch.get_cert('leaf-aa'),
    ]
    signer = signers.SimpleSigner(
        signing_cert=pki_arch.get_cert(CertLabel('signer1')),
        signing_key=pki_arch.key_set.get_private_key(KeyLabel('signer1')),
        cert_registry=SimpleCertificateStore.from_certs(authorities),
        attribute_certs=[
            pki_arch.get_attr_cert(CertLabel('alice-role-with-rev'))
        ],
    )
    dummy_ts = timestamps.DummyTimeStamper(
        tsa_cert=pki_arch.get_cert(CertLabel('tsa')),
        tsa_key=pki_arch.key_set.get_private_key(KeyLabel('tsa')),
        certs_to_embed=SimpleCertificateStore.from_certs(
            [pki_arch.get_cert('root')]
        ),
    )

    fetchers = RequestsFetcherBackend().get_fetchers()
    vc = ValidationContext(
        trust_roots=[pki_arch.get_cert('root')],
        allow_fetching=True,
        other_certs=authorities,
        fetchers=fetchers,
        revocation_mode='require',
    )
    ac_vc = ValidationContext(
        trust_roots=[pki_arch.get_cert('root-aa')],
        allow_fetching=True,
        other_certs=authorities,
        fetchers=fetchers,
        revocation_mode='require',
    )
    Illusionist(pki_arch).register(requests_mock)

    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            validation_context=vc,
            ac_validation_context=ac_vc,
            subfilter=PADES,
            embed_validation_info=True,
            dss_settings=DSSContentSettings(
                include_vri=False,
                placement=SigDSSPlacementPreference.TOGETHER_WITH_SIGNATURE,
            ),
        ),
        signer=signer,
        timestamper=dummy_ts,
        existing_fields_only=True,
    )

    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    # 4 CA certs, 1 AA certs, 1 AC, 1 signer cert, 1 TSA cert,
    # 1 OCSP responder cert -> 9 certs total
    # 7 (4 CA + 1 AA + 1 AC + 1 signer) are in the CMS payload
    assert len(s.other_embedded_certs) == 5
    assert len(s.embedded_attr_certs) == 1
    dss = r.root['/DSS']
    assert len(dss['/Certs']) == 8  # no ACs here, but OCSP and TSA are present
    assert len(dss['/OCSPs']) == 2  # signer + AC (leaf-aa has OCSP)
    assert len(dss['/CRLs']) == 3  # root, interm-aa, root-aa
    status = await async_validate_pdf_ltv_signature(
        s,
        validation_type=RevocationInfoValidationType.PADES_LT,
        validation_context_kwargs={
            'trust_roots': [pki_arch.get_cert('root')],
            'revocation_mode': 'require',
        },
        ac_validation_context_kwargs={
            'trust_roots': [pki_arch.get_cert('root-aa')],
            'revocation_mode': 'require',
        },
    )
    assert status.bottom_line
    roles = list(status.ac_attrs['role'].attr_values)
    role = roles[0]
    assert isinstance(role, cms.RoleSyntax)
    assert len(status.ac_attrs) == 1
    assert role['role_name'].native == 'bigboss@example.com'


@pytest.mark.parametrize('with_force_revinfo', [True, False])
@pytest.mark.asyncio
async def test_pades_lta_live_ac_presign_validation(
    requests_mock, with_force_revinfo
):
    # Same as the above, but with LTA instead (+some time manipulation)

    with freeze_time('2020-11-01'):
        pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))
        authorities = [
            pki_arch.get_cert('root'),
            pki_arch.get_cert('interm'),
            pki_arch.get_cert('root-aa'),
            pki_arch.get_cert('interm-aa'),
            pki_arch.get_cert('leaf-aa'),
        ]
        signer = signers.SimpleSigner(
            signing_cert=pki_arch.get_cert(CertLabel('signer1')),
            signing_key=pki_arch.key_set.get_private_key(KeyLabel('signer1')),
            cert_registry=SimpleCertificateStore.from_certs(authorities),
            attribute_certs=[
                pki_arch.get_attr_cert(CertLabel('alice-role-with-rev'))
            ],
        )
        dummy_ts = timestamps.DummyTimeStamper(
            tsa_cert=pki_arch.get_cert(CertLabel('tsa')),
            tsa_key=pki_arch.key_set.get_private_key(KeyLabel('tsa')),
            certs_to_embed=SimpleCertificateStore.from_certs(
                [pki_arch.get_cert('root')]
            ),
        )

        fetchers = RequestsFetcherBackend().get_fetchers()
        vc = ValidationContext(
            trust_roots=[pki_arch.get_cert('root')],
            allow_fetching=True,
            other_certs=authorities,
            fetchers=fetchers,
            revocation_mode='require',
        )
        ac_vc = ValidationContext(
            trust_roots=[pki_arch.get_cert('root-aa')],
            allow_fetching=True,
            other_certs=authorities,
            fetchers=fetchers,
            revocation_mode='require',
        )
        Illusionist(pki_arch).register(requests_mock)

        w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
        out = await signers.async_sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                validation_context=vc,
                ac_validation_context=ac_vc,
                subfilter=PADES,
                embed_validation_info=True,
                use_pades_lta=True,
            ),
            signer=signer,
            timestamper=dummy_ts,
            existing_fields_only=True,
        )

    revo_policy = CertRevTrustPolicy(
        RevocationCheckingPolicy(
            RevocationCheckingRule.CRL_OR_OCSP_REQUIRED,
            RevocationCheckingRule.CRL_OR_OCSP_REQUIRED,
        )
    )

    with freeze_time('2028-02-01'):
        r = PdfFileReader(out)
        s = r.embedded_signatures[0]
        vc_kwargs = {
            'trust_roots': [pki_arch.get_cert('root')],
        }
        ac_vc_kwargs = {
            'trust_roots': [pki_arch.get_cert('root-aa')],
        }
        if not with_force_revinfo:
            # supply parameters the usual way
            vc_kwargs['revinfo_policy'] = revo_policy
            ac_vc_kwargs['revinfo_policy'] = revo_policy
        status = await async_validate_pdf_ltv_signature(
            s,
            validation_type=RevocationInfoValidationType.PADES_LTA,
            validation_context_kwargs=vc_kwargs,
            ac_validation_context_kwargs=ac_vc_kwargs,
            force_revinfo=with_force_revinfo,
        )
        assert status.bottom_line
        roles = list(status.ac_attrs['role'].attr_values)
        role = roles[0]
        assert isinstance(role, cms.RoleSyntax)
        assert len(status.ac_attrs) == 1
        assert role['role_name'].native == 'bigboss@example.com'


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_cades_signer_attrs_autofill_dss(requests_mock):
    pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))
    signer = signers.SimpleSigner(
        signing_cert=pki_arch.get_cert(CertLabel('signer1')),
        signing_key=pki_arch.key_set.get_private_key(KeyLabel('signer1')),
        cert_registry=SimpleCertificateStore(),  # no certs here on purpose
    )
    main_vc, ac_vc = live_ac_vcs(requests_mock, with_authorities=True)
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            embed_validation_info=True,
            validation_context=main_vc,
            ac_validation_context=ac_vc,
            cades_signed_attr_spec=CAdESSignedAttrSpec(
                commitment_type=CommitmentTypeIndication(
                    {
                        'commitment_type_id': 'proof_of_approval',
                    }
                ),
                signer_attributes=SignerAttrSpec(
                    claimed_attrs=[SAMPLE_GROUP_ATTR],
                    certified_attrs=[
                        pki_arch.get_attr_cert(CertLabel('alice-role-with-rev'))
                    ],
                ),
            ),
        ),
        signer=signer,
    )
    r = PdfFileReader(out)
    # 4 CA certs, 1 AA certs, 1 signer cert, 1 OCSP responder cert -> 7 certs
    dss = r.root['/DSS']
    assert len(dss['/Certs']) == 7


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_cades_signer_attrs_validate_acs(requests_mock):
    pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))
    signer = signers.SimpleSigner(
        signing_cert=pki_arch.get_cert(CertLabel('signer1')),
        signing_key=pki_arch.key_set.get_private_key(KeyLabel('signer1')),
        cert_registry=SimpleCertificateStore.from_certs(
            [
                pki_arch.get_cert('root'),
                pki_arch.get_cert('interm'),
                pki_arch.get_cert('root-aa'),
                pki_arch.get_cert('interm-aa'),
                pki_arch.get_cert('leaf-aa'),
            ]
        ),
    )
    main_vc, ac_vc = live_ac_vcs(requests_mock, with_authorities=True)
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            embed_validation_info=True,
            validation_context=main_vc,
            ac_validation_context=ac_vc,
            cades_signed_attr_spec=CAdESSignedAttrSpec(
                commitment_type=CommitmentTypeIndication(
                    {
                        'commitment_type_id': 'proof_of_approval',
                    }
                ),
                signer_attributes=SignerAttrSpec(
                    claimed_attrs=[SAMPLE_GROUP_ATTR],
                    certified_attrs=[
                        pki_arch.get_attr_cert(CertLabel('alice-role-with-rev'))
                    ],
                ),
            ),
        ),
        signer=signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert len(s.embedded_attr_certs) == 0  # nothing here
    main_vc, ac_vc = live_ac_vcs(requests_mock)
    status = await async_validate_pdf_signature(
        s, signer_validation_context=main_vc, ac_validation_context=ac_vc
    )
    assert status.bottom_line
    # this one was only 'claimed'
    assert 'group' not in status.ac_attrs
    roles = list(status.ac_attrs['role'].attr_values)
    role = roles[0]
    assert isinstance(role, cms.RoleSyntax)
    assert role['role_name'].native == 'bigboss@example.com'

    # also perform checks for the CAdES signer attrs info
    cades_signer_attrs = status.cades_signer_attrs
    assert cades_signer_attrs is not None
    assert len(cades_signer_attrs.claimed_attrs) == 1
    assert len(cades_signer_attrs.certified_attrs) == 1

    # claimed attrs
    assert 'role' not in cades_signer_attrs.claimed_attrs
    (all_values,) = iter(cades_signer_attrs.claimed_attrs)
    assert all_values == cades_signer_attrs.claimed_attrs['group']
    (groups_ietf_attr,) = iter(all_values.attr_values)
    assert isinstance(groups_ietf_attr, cms.IetfAttrSyntax)
    groups = groups_ietf_attr['values']
    assert set(groups.native) == {'Executives', 'Employees'}

    # certified attrs
    assert 'group' not in cades_signer_attrs.certified_attrs
    roles = list(cades_signer_attrs.certified_attrs['role'].attr_values)
    role = roles[0]
    assert isinstance(role, cms.RoleSyntax)
    assert role['role_name'].native == 'bigboss@example.com'


@freeze_time('2020-11-01')
@pytest.mark.parametrize('pass_ac_vc', [True, False])
@pytest.mark.asyncio
async def test_cades_signer_attrs_claimed_only(requests_mock, pass_ac_vc):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            cades_signed_attr_spec=CAdESSignedAttrSpec(
                commitment_type=CommitmentTypeIndication(
                    {
                        'commitment_type_id': 'proof_of_approval',
                    }
                ),
                signer_attributes=SignerAttrSpec(
                    claimed_attrs=[SAMPLE_GROUP_ATTR], certified_attrs=()
                ),
            ),
        ),
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert len(s.embedded_attr_certs) == 0  # nothing here
    main_vc = live_testing_vc(requests_mock)
    status = await async_validate_pdf_signature(
        s,
        signer_validation_context=main_vc,
        # what the VC contains shouldn't matter, since there are no ACs to
        # validate
        ac_validation_context=(main_vc if pass_ac_vc else None),
    )
    assert status.bottom_line
    assert not status.ac_attrs
    if not pass_ac_vc:
        assert status.ac_attrs is None

    cades_signer_attrs = status.cades_signer_attrs
    assert cades_signer_attrs is not None
    assert len(cades_signer_attrs.claimed_attrs) == 1
    assert not cades_signer_attrs.certified_attrs
    if not pass_ac_vc:
        assert cades_signer_attrs.certified_attrs is None

    # claimed attrs
    (groups_ietf_attr,) = iter(
        cades_signer_attrs.claimed_attrs['group'].attr_values
    )
    assert isinstance(groups_ietf_attr, cms.IetfAttrSyntax)
    groups = groups_ietf_attr['values']
    assert set(groups.native) == {'Executives', 'Employees'}


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_cades_signer_attrs_validate_acs_no_claimed(requests_mock):
    pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))
    signer = signers.SimpleSigner(
        signing_cert=pki_arch.get_cert(CertLabel('signer1')),
        signing_key=pki_arch.key_set.get_private_key(KeyLabel('signer1')),
        cert_registry=SimpleCertificateStore.from_certs(
            [
                pki_arch.get_cert('root'),
                pki_arch.get_cert('interm'),
                pki_arch.get_cert('root-aa'),
                pki_arch.get_cert('interm-aa'),
                pki_arch.get_cert('leaf-aa'),
            ]
        ),
    )
    main_vc, ac_vc = live_ac_vcs(requests_mock, with_authorities=True)
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            embed_validation_info=True,
            validation_context=main_vc,
            ac_validation_context=ac_vc,
            cades_signed_attr_spec=CAdESSignedAttrSpec(
                signer_attributes=SignerAttrSpec(
                    claimed_attrs=(),
                    certified_attrs=[
                        pki_arch.get_attr_cert(CertLabel('alice-role-with-rev'))
                    ],
                )
            ),
        ),
        signer=signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert len(s.embedded_attr_certs) == 0  # nothing here
    main_vc, ac_vc = live_ac_vcs(requests_mock)
    status = await async_validate_pdf_signature(
        s, signer_validation_context=main_vc, ac_validation_context=ac_vc
    )
    assert status.bottom_line
    assert 'CERTIFIED_SIGNER_ATTRS_INVALID' not in status.summary()
    # this one was only 'claimed'
    roles = list(status.ac_attrs['role'].attr_values)
    role = roles[0]
    assert isinstance(role, cms.RoleSyntax)
    assert role['role_name'].native == 'bigboss@example.com'

    # also perform checks for the CAdES signer attrs info
    cades_signer_attrs = status.cades_signer_attrs
    assert cades_signer_attrs is not None
    assert not cades_signer_attrs.claimed_attrs
    assert len(cades_signer_attrs.certified_attrs) == 1

    roles = list(cades_signer_attrs.certified_attrs['role'].attr_values)
    role = roles[0]
    assert isinstance(role, cms.RoleSyntax)
    assert role['role_name'].native == 'bigboss@example.com'
    assert not cades_signer_attrs.unknown_attrs_present


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_cades_signer_attrs_validate_acs_wrong_vc(requests_mock):
    pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))
    signer = signers.SimpleSigner(
        signing_cert=pki_arch.get_cert(CertLabel('signer1')),
        signing_key=pki_arch.key_set.get_private_key(KeyLabel('signer1')),
        cert_registry=SimpleCertificateStore.from_certs(
            [
                pki_arch.get_cert('root'),
                pki_arch.get_cert('interm'),
                pki_arch.get_cert('root-aa'),
                pki_arch.get_cert('interm-aa'),
                pki_arch.get_cert('leaf-aa'),
            ]
        ),
    )
    main_vc, ac_vc = live_ac_vcs(requests_mock, with_authorities=True)
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            embed_validation_info=True,
            validation_context=main_vc,
            ac_validation_context=ac_vc,
            cades_signed_attr_spec=CAdESSignedAttrSpec(
                commitment_type=CommitmentTypeIndication(
                    {
                        'commitment_type_id': 'proof_of_approval',
                    }
                ),
                signer_attributes=SignerAttrSpec(
                    claimed_attrs=[SAMPLE_GROUP_ATTR],
                    certified_attrs=[
                        pki_arch.get_attr_cert(CertLabel('alice-role-with-rev'))
                    ],
                ),
            ),
        ),
        signer=signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert len(s.embedded_attr_certs) == 0  # nothing here
    main_vc, ac_vc = live_ac_vcs(requests_mock)

    # validate with the wrong AC VC
    status = await async_validate_pdf_signature(
        s, signer_validation_context=main_vc, ac_validation_context=main_vc
    )

    assert not status.cades_signer_attrs.valid
    assert 'CERTIFIED_SIGNER_ATTRS_INVALID' in status.summary()

    assert len(status.ac_attrs) == 0

    # also perform checks for the CAdES signer attrs info
    cades_signer_attrs = status.cades_signer_attrs
    assert cades_signer_attrs is not None
    assert len(cades_signer_attrs.claimed_attrs) == 1
    assert len(cades_signer_attrs.certified_attrs) == 0
    assert len(status.ac_validation_errs) == 1
    assert len(cades_signer_attrs.ac_validation_errs) == 1
    assert not cades_signer_attrs.unknown_attrs_present


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_cades_signer_attrs_validate_acs_missing_vc(requests_mock):
    pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))
    signer = signers.SimpleSigner(
        signing_cert=pki_arch.get_cert(CertLabel('signer1')),
        signing_key=pki_arch.key_set.get_private_key(KeyLabel('signer1')),
        cert_registry=SimpleCertificateStore.from_certs(
            [
                pki_arch.get_cert('root'),
                pki_arch.get_cert('interm'),
                pki_arch.get_cert('root-aa'),
                pki_arch.get_cert('interm-aa'),
                pki_arch.get_cert('leaf-aa'),
            ]
        ),
    )
    main_vc, ac_vc = live_ac_vcs(requests_mock, with_authorities=True)
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
            embed_validation_info=True,
            validation_context=main_vc,
            ac_validation_context=ac_vc,
            cades_signed_attr_spec=CAdESSignedAttrSpec(
                commitment_type=CommitmentTypeIndication(
                    {
                        'commitment_type_id': 'proof_of_approval',
                    }
                ),
                signer_attributes=SignerAttrSpec(
                    claimed_attrs=[SAMPLE_GROUP_ATTR],
                    certified_attrs=[
                        pki_arch.get_attr_cert(CertLabel('alice-role-with-rev'))
                    ],
                ),
            ),
        ),
        signer=signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert len(s.embedded_attr_certs) == 0  # nothing here
    main_vc, _ = live_ac_vcs(requests_mock)

    # validate with the wrong AC VC
    status = await async_validate_pdf_signature(
        s,
        signer_validation_context=main_vc,
    )

    # this should _not_ fail validation, since we've indicated that we don't
    # care about ACs
    assert status.bottom_line
    assert 'CERTIFIED_SIGNER_ATTRS_INVALID' not in status.summary()
    assert status.ac_attrs is None

    cades_signer_attrs = status.cades_signer_attrs
    assert cades_signer_attrs is not None
    assert len(cades_signer_attrs.claimed_attrs) == 1
    assert cades_signer_attrs.certified_attrs is None
    assert status.ac_validation_errs is None
    assert cades_signer_attrs.ac_validation_errs is None


@freeze_time('2020-11-01')
@pytest.mark.parametrize(
    'as_signed_assertions,pass_ac_vc',
    [(True, True), (True, False), (False, True), (False, False)],
)
@pytest.mark.asyncio
async def test_cades_signer_attrs_unknown_attrs(
    requests_mock, as_signed_assertions, pass_ac_vc
):
    class CustomAttrProvider(CMSAttributeProvider):
        attribute_type = 'signer_attributes_v2'

        async def build_attr_value(self, dry_run=False):
            signer_attrs = {
                'claimed_attributes': [SAMPLE_GROUP_ATTR],
            }
            value = core.OctetString(b'\xde\xad\xbe\xef')
            if as_signed_assertions:
                signer_attrs['signed_assertions'] = [
                    SignedAssertion(
                        {
                            'signed_assertion_id': '2.999',
                            'signed_assertion': value,
                        }
                    )
                ]
            else:
                signer_attrs['certified_attributes_v2'] = [
                    CertifiedAttributeChoices(
                        name='other_attr_cert',
                        value={
                            'other_attr_cert_id': '2.999',
                            'other_attr_cert': value,
                        },
                    )
                ]
            return SignerAttributesV2(signer_attrs)

    class CustomSigner(signers.SimpleSigner):
        def _signed_attr_providers(self, *args, **kwargs):
            yield from super()._signed_attr_providers(*args, **kwargs)
            yield CustomAttrProvider()

    signer = CustomSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
        ),
        signer=signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert len(s.embedded_attr_certs) == 0  # nothing here
    main_vc = live_testing_vc(requests_mock)
    status = await async_validate_pdf_signature(
        s,
        signer_validation_context=main_vc,
        # what the VC contains shouldn't matter, since there are no ACs to
        # validate
        ac_validation_context=(main_vc if pass_ac_vc else None),
    )
    # bottom line is unaffected
    assert status.bottom_line
    assert not status.ac_attrs
    if not pass_ac_vc:
        assert status.ac_attrs is None

    cades_signer_attrs = status.cades_signer_attrs
    assert cades_signer_attrs is not None
    # check for unknown_attrs_present
    assert cades_signer_attrs.unknown_attrs_present
    assert len(cades_signer_attrs.claimed_attrs) == 1
    assert not cades_signer_attrs.certified_attrs
    if not pass_ac_vc:
        assert cades_signer_attrs.certified_attrs is None

    # claimed attrs should still be OK
    (groups_ietf_attr,) = iter(
        cades_signer_attrs.claimed_attrs['group'].attr_values
    )
    assert isinstance(groups_ietf_attr, cms.IetfAttrSyntax)
    groups = groups_ietf_attr['values']
    assert set(groups.native) == {'Executives', 'Employees'}


@freeze_time('2020-11-01')
@pytest.mark.parametrize('pass_ac_vc', [True, False])
@pytest.mark.asyncio
async def test_cades_signer_attrs_multivalued(requests_mock, pass_ac_vc):
    class CustomSigner(signers.SimpleSigner):
        async def signed_attrs(self, *args, **kwargs):
            signed_attrs = await super().signed_attrs(*args, **kwargs)
            signer_attrs = {
                'claimed_attributes': [SAMPLE_GROUP_ATTR],
            }
            attr = SignerAttributesV2(signer_attrs)
            signed_attrs.append(
                cms.CMSAttribute(
                    {'type': 'signer_attributes_v2', 'values': [attr, attr]}
                )
            )
            return signed_attrs

    signer = CustomSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            subfilter=PADES,
        ),
        signer=signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert len(s.embedded_attr_certs) == 0  # nothing here
    main_vc = live_testing_vc(requests_mock)
    with pytest.raises(SignatureValidationError, match='Expected single'):
        await async_validate_pdf_signature(
            s,
            signer_validation_context=main_vc,
            # what the VC contains shouldn't matter, since there are no ACs to
            # validate
            ac_validation_context=(main_vc if pass_ac_vc else None),
        )


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_interrupted_nonstrict_with_psi():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_SLIGHTLY_BROKEN), strict=False)
    pdf_signer = signers.PdfSigner(
        signers.PdfSignatureMetadata(
            field_name='SigNew',
            subfilter=PADES,
            embed_validation_info=True,
            validation_context=SIMPLE_V_CONTEXT(),
        ),
        signer=FROM_CA,
        timestamper=DUMMY_TS,
    )
    prep_digest, tbs_document, output = (
        await pdf_signer.async_digest_doc_for_signing(w)
    )
    md_algorithm = tbs_document.md_algorithm
    assert tbs_document.post_sign_instructions is not None

    await PdfTBSDocument.async_finish_signing(
        output,
        prep_digest,
        await FROM_CA.async_sign(
            prep_digest.document_digest,
            digest_algorithm=md_algorithm,
        ),
        post_sign_instr=tbs_document.post_sign_instructions,
    )

    r = PdfFileReader(output, strict=False)
    await async_val_trusted(r.embedded_signatures[0], extd=True)



"""
Utilities for stamping PDF files.

Here 'stamping' loosely refers to adding small overlays (QR codes, text boxes,
etc.) on top of already existing content in PDF files.

The code in this module is also used by the :mod:`.sign` module to render
signature appearances.
"""

import enum
import uuid
from binascii import hexlify
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Tuple

import qrcode
import tzlocal
from pyhanko.config.api import ConfigurableMixin
from pyhanko.config.errors import ConfigurationError
from pyhanko.pdf_utils import content, generic, layout
from pyhanko.pdf_utils.generic import IndirectObject, pdf_name, pdf_string
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.layout import LayoutError
from pyhanko.pdf_utils.misc import rd
from pyhanko.pdf_utils.qr import PdfFancyQRImage, PdfStreamQRImage
from pyhanko.pdf_utils.text import DEFAULT_BOX_LAYOUT, TextBox, TextBoxStyle
from pyhanko.pdf_utils.writer import BasePdfFileWriter, init_xobject_dictionary

__all__ = [
    "AnnotAppearances",
    "BaseStampStyle",
    "TextStampStyle",
    "QRStampStyle",
    "StaticStampStyle",
    "QRPosition",
    "CoordinateSystem",
    "BaseStamp",
    "TextStamp",
    "QRStamp",
    "StaticContentStamp",
    "text_stamp_file",
    "qr_stamp_file",
    "STAMP_ART_CONTENT",
]


class AnnotAppearances:
    """
    Convenience abstraction to set up an appearance dictionary for a PDF
    annotation.

    Annotations can have three appearance streams, which can be roughly
    characterised as follows:

    * *normal*: the only required one, and the default one;
    * *rollover*: used when mousing over the annotation;
    * *down*: used when clicking the annotation.

    These are given as references to form XObjects.

    .. note::
        This class only covers the simple case of an appearance dictionary
        for an annotation with only one appearance state.

    See § 12.5.5 in ISO 32000-1 for further information.
    """

    def __init__(
        self,
        normal: generic.IndirectObject,
        rollover: Optional[generic.IndirectObject] = None,
        down: Optional[generic.IndirectObject] = None,
    ):
        self.normal = normal
        self.rollover = rollover
        self.down = down

    def as_pdf_object(self) -> generic.DictionaryObject:
        """
        Convert the :class:`.AnnotationAppearances` instance to a PDF
        dictionary.

        :return:
            A :class:`~.pdf_utils.generic.DictionaryObject` that can be plugged
            into the ``/AP`` entry of an annotation dictionary.
        """

        res = generic.DictionaryObject({pdf_name('/N'): self.normal})
        if self.rollover is not None:
            res[pdf_name('/R')] = self.rollover
        if self.down is not None:
            res[pdf_name('/D')] = self.down
        return res


def _get_background_content(bg_spec) -> content.PdfContent:
    if not isinstance(bg_spec, str):
        raise ConfigurationError("Background specification must be a string")
    # 'special' value to use the stamp vector image baked into
    # the module
    if bg_spec == '__stamp__':
        return STAMP_ART_CONTENT
    elif bg_spec.endswith('.pdf'):
        # import first page of PDF as background
        return content.ImportedPdfPage(bg_spec)
    else:
        from PIL import Image
        from pyhanko.pdf_utils.images import PdfImage

        img = Image.open(bg_spec)
        # Setting the writer can be delayed
        return PdfImage(img, writer=None)


@dataclass(frozen=True)
class BaseStampStyle(ConfigurableMixin):
    """
    Base class for stamp styles.
    """

    border_width: int = 3
    """
    Border width in user units (for the stamp, not the text box).
    """

    background: Optional[content.PdfContent] = None
    """
    :class:`~.pdf_utils.content.PdfContent` instance that will be used to render
    the stamp's background.
    """

    background_layout: layout.SimpleBoxLayoutRule = layout.SimpleBoxLayoutRule(
        x_align=layout.AxisAlignment.ALIGN_MID,
        y_align=layout.AxisAlignment.ALIGN_MID,
        margins=layout.Margins.uniform(5),
    )
    """
    Layout rule to render the background inside the stamp's bounding box.
    Only used if the background has a fully specified :attr:`PdfContent.box`.

    Otherwise, the renderer will position the cursor at
    ``(left_margin, bottom_margin)`` and render the content as-is.
    """

    background_opacity: float = 0.6
    """
    Opacity value to render the background at. This should be a floating-point
    number between `0` and `1`.
    """

    @classmethod
    def process_entries(cls, config_dict):
        """
        This implementation of :meth:`process_entries` processes the
        :attr:`background` configuration value.
        This can either be a path to an image file, in which case it will
        be turned into an instance of :class:`~.pdf_utils.images.PdfImage`,
        or the special value ``__stamp__``, which is an alias for
        :const:`~pyhanko.stamp.STAMP_ART_CONTENT`.
        """

        super().process_entries(config_dict)
        bg_spec = None
        try:
            bg_spec = config_dict['background']
        except KeyError:
            pass
        if bg_spec is not None:
            config_dict['background'] = _get_background_content(bg_spec)

    def create_stamp(
        self,
        writer: BasePdfFileWriter,
        box: layout.BoxConstraints,
        text_params: dict,
    ) -> 'BaseStamp':
        raise NotImplementedError


@dataclass(frozen=True)
class StaticStampStyle(BaseStampStyle):
    """
    Stamp style that does not include any custom parts; it only renders
    the background.
    """

    background_opacity: float = 1.0
    """
    Opacity value to render the background at. This should be a floating-point
    number between `0` and `1`.
    """

    @classmethod
    def from_pdf_file(
        cls, file_name, page_ix=0, **kwargs
    ) -> 'StaticStampStyle':
        """
        Create a :class:`StaticStampStyle` from a page from an external PDF
        document. This is a convenience wrapper around
        :class:`~content.ImportedPdfContent`.

        The remaining keyword arguments are passed to
        :class:`StaticStampStyle`'s init method.

        :param file_name:
            File name of the external PDF document.
        :param page_ix:
            Page index to import. The default is ``0``, i.e. the first page.
        """
        return StaticStampStyle(
            background=content.ImportedPdfPage(file_name, page_ix=page_ix),
            **kwargs,
        )

    def create_stamp(
        self,
        writer: BasePdfFileWriter,
        box: layout.BoxConstraints,
        text_params: dict,
    ) -> 'StaticContentStamp':
        return StaticContentStamp(writer=writer, style=self, box=box)


@dataclass(frozen=True)
class TextStampStyle(BaseStampStyle):
    """
    Style for text-based stamps.

    Roughly speaking, this stamp type renders some predefined (but parametrised)
    piece of text inside a text box, and possibly applies a background to it.
    """

    text_box_style: TextBoxStyle = TextBoxStyle()
    """
    The text box style for the internal text box used.
    """

    inner_content_layout: Optional[layout.SimpleBoxLayoutRule] = None
    """
    Rule determining the position and alignment of the inner text box within
    the stamp.
    
    .. warning::
        This only affects the position of the box, not the alignment of the
        text within.
    """

    stamp_text: str = '%(ts)s'
    """
    Text template for the stamp. The template can contain an interpolation
    parameter ``ts`` that will be replaced by the stamping time.
    
    Additional parameters may be added if necessary. Values for these must be
    passed to the :meth:`~.TextStamp.__init__` method of the 
    :class:`.TextStamp` class in the ``text_params`` argument.
    """

    timestamp_format: str = '%Y-%m-%d %H:%M:%S %Z'
    """
    Datetime format used to render the timestamp.
    """

    def create_stamp(
        self,
        writer: BasePdfFileWriter,
        box: layout.BoxConstraints,
        text_params: dict,
    ) -> 'TextStamp':
        return TextStamp(
            writer=writer, style=self, box=box, text_params=text_params
        )


class QRPosition(enum.Enum):
    """
    QR positioning constants, with the corresponding default content layout
    rule.
    """

    LEFT_OF_TEXT = layout.SimpleBoxLayoutRule(
        x_align=layout.AxisAlignment.ALIGN_MIN,
        y_align=layout.AxisAlignment.ALIGN_MID,
    )
    RIGHT_OF_TEXT = layout.SimpleBoxLayoutRule(
        x_align=layout.AxisAlignment.ALIGN_MAX,
        y_align=layout.AxisAlignment.ALIGN_MID,
    )
    ABOVE_TEXT = layout.SimpleBoxLayoutRule(
        y_align=layout.AxisAlignment.ALIGN_MAX,
        x_align=layout.AxisAlignment.ALIGN_MID,
    )
    BELOW_TEXT = layout.SimpleBoxLayoutRule(
        y_align=layout.AxisAlignment.ALIGN_MIN,
        x_align=layout.AxisAlignment.ALIGN_MID,
    )

    @property
    def horizontal_flow(self):
        return self in (QRPosition.LEFT_OF_TEXT, QRPosition.RIGHT_OF_TEXT)

    @classmethod
    def from_config(cls, config_str) -> 'QRPosition':
        """
        Convert from a configuration string.

        :param config_str:
            A string: 'left', 'right', 'top', 'bottom'
        :return:
            An :class:`.QRPosition` value.
        :raise ConfigurationError: on unexpected string inputs.
        """
        try:
            return {
                'left': QRPosition.LEFT_OF_TEXT,
                'right': QRPosition.RIGHT_OF_TEXT,
                'top': QRPosition.ABOVE_TEXT,
                'bottom': QRPosition.BELOW_TEXT,
            }[config_str.lower()]
        except KeyError:
            raise ConfigurationError(
                f"'{config_str}' is not a valid QR position setting; valid "
                f"values are 'left', 'right', 'top', 'bottom'"
            )


DEFAULT_QR_SCALE = 0.2
"""
If the layout & other bounding boxes don't impose another size requirement,
render QR codes at ~20% of their natural size in QR canvas units. At scale 1,
this produces codes of about 2cm x 2cm for a 25-module QR code, which is
probably OK.
"""


@dataclass(frozen=True)
class QRStampStyle(TextStampStyle):
    """
    Style for text-based stamps together with a QR code.

    This is exactly the same as a text stamp, except that the text box
    is rendered with a QR code to the left of it.
    """

    innsep: int = 3
    """
    Inner separation inside the stamp.
    """

    stamp_text: str = (
        "Digital version available at\n"
        "this url: %(url)s\n"
        "Timestamp: %(ts)s"
    )
    """
    Text template for the stamp.
    The description of :attr:`.TextStampStyle.stamp_text` still applies, but
    an additional default interpolation parameter ``url`` is available.
    This parameter will be replaced with the URL that the QR code points to.
    """

    qr_inner_size: Optional[int] = None
    """
    Size of the QR code in the inner layout. By default, this is in user units,
    but if the stamp has a fully defined bounding box, it may be rescaled
    depending on :attr:`inner_content_layout`.
    
    If unspecified, a reasonable default will be used.
    """

    qr_position: QRPosition = QRPosition.LEFT_OF_TEXT
    """
    Position of the QR code relative to the text box.
    """

    qr_inner_content: Optional[content.PdfContent] = None
    """
    Inner graphics content to be included in the QR code (experimental).
    """

    @classmethod
    def process_entries(cls, config_dict):
        super().process_entries(config_dict)

        try:
            qr_pos = config_dict['qr_position']
            config_dict['qr_position'] = QRPosition.from_config(qr_pos)
        except KeyError:
            pass

    def create_stamp(
        self,
        writer: BasePdfFileWriter,
        box: layout.BoxConstraints,
        text_params: dict,
    ) -> 'QRStamp':
        # extract the URL parameter
        try:
            url = text_params.pop('url')
        except KeyError:
            raise layout.LayoutError(
                "Using a QR stamp style requires a 'url' text parameter."
            )
        return QRStamp(
            writer, style=self, url=url, text_params=text_params, box=box
        )


class CoordinateSystem(enum.Enum):
    """
    Positioning convention for stamps.
    """

    PAGE_DEFAULT = 1
    """
    Always treat the stamp's position in the page's default coordinate system,
    by defensively forcing a restore to the original graphics state.

    .. note::
        This is the default behaviour since ``0.27.0``.
    """

    AMBIENT = 2
    """
    Apply the stamp in the ambient frame of reference set by the existing
    page content.
    This may yield unpredictable results depending on the input document.

    .. note::
        This was the default behaviour prior to ``0.27.0``.
    """


class BaseStamp(content.PdfContent):
    def __init__(
        self,
        writer: BasePdfFileWriter,
        style,
        box: Optional[layout.BoxConstraints] = None,
    ):
        super().__init__(box=box, writer=writer)
        self.style = style
        self._resources_ready = False
        self._stamp_ref: Optional[IndirectObject] = None

    def _render_background(self):
        bg = self.style.background
        bg.set_writer(self.writer)
        bg_content = bg.render()  # render first, in case the BBox is lazy

        bg_box = bg.box
        if bg_box.width_defined and bg_box.height_defined:
            # apply layout rule
            positioning = self.style.background_layout.fit(
                self.box, bg_box.width, bg_box.height
            )
        else:
            # No idea about the background dimensions, so just use
            # the left/bottom margins and hope for the best
            margins = self.style.background_layout.margins
            positioning = layout.Positioning(
                x_scale=1, y_scale=1, x_pos=margins.left, y_pos=margins.bottom
            )

        # set opacity in graphics state
        opacity = generic.FloatObject(self.style.background_opacity)
        self.set_resource(
            category=content.ResourceType.EXT_G_STATE,
            name=pdf_name('/BackgroundGS'),
            value=generic.DictionaryObject(
                {pdf_name('/CA'): opacity, pdf_name('/ca'): opacity}
            ),
        )

        # Position & render the background
        command = b'q /BackgroundGS gs %s %s Q' % (
            positioning.as_cm(),
            bg_content,
        )
        # we do this after render(), just in case our background resource
        # decides to pull in extra stuff during rendering
        self.import_resources(bg.resources)
        return command

    def _render_inner_content(self):
        raise NotImplementedError

    def render(self):
        command_stream = [b'q']

        inner_content = self._render_inner_content()

        # Now that the inner layout is done, the dimensions of our bounding
        # box should all have been reified. Let's put in the background,
        # if there is one
        if self.style.background:
            command_stream.append(self._render_background())

        # put in the inner content
        if inner_content:
            command_stream.extend(inner_content)

        # draw the border around the stamp
        bbox = self.box
        border_width = self.style.border_width
        if border_width:
            command_stream.append(
                b'%g w 0 0 %g %g re S' % (border_width, bbox.width, bbox.height)
            )
        command_stream.append(b'Q')
        return b' '.join(command_stream)

    def register(self) -> generic.IndirectObject:
        """
        Register the stamp with the writer coupled to this instance, and
        cache the returned reference.

        This works by calling :meth:`.PdfContent.as_form_xobject`.

        :return:
            An indirect reference to the form XObject containing the stamp.
        """
        stamp_ref = self._stamp_ref
        if stamp_ref is None:
            wr = self._ensure_writer
            form_xobj = self.as_form_xobject()
            self._stamp_ref = stamp_ref = wr.add_object(form_xobj)
        return stamp_ref

    def apply(
        self,
        dest_page: int,
        x: int,
        y: int,
        *,
        coords: CoordinateSystem = CoordinateSystem.PAGE_DEFAULT,
    ):
        """
        Apply a stamp to a particular page in the PDF writer attached to this
        :class:`.BaseStamp` instance.

        :param dest_page:
            Index of the page to which the stamp is to be applied
            (starting at `0`).
        :param x:
            Horizontal position of the stamp's lower left corner on the page.
        :param y:
            Vertical position of the stamp's lower left corner on the page.
        :param coords:
            The coordinate convention to use.
        :return:
            A reference to the affected page object, together with
            a ``(width, height)`` tuple describing the dimensions of the stamp.
        """
        stamp_ref = self.register()
        resource_name = b'/Stamp' + hexlify(uuid.uuid4().bytes)
        stamp_paint = b'q 1 0 0 1 %g %g cm %s Do Q' % (
            rd(x),
            rd(y),
            resource_name,
        )
        stamp_wrapper_stream = generic.StreamObject(stream_data=stamp_paint)
        resources = generic.DictionaryObject(
            {
                pdf_name('/XObject'): generic.DictionaryObject(
                    {pdf_name(resource_name.decode('ascii')): stamp_ref}
                )
            }
        )
        wr = self.writer
        assert wr is not None
        if coords == CoordinateSystem.PAGE_DEFAULT:
            # wrap the existing content in q / Q
            wr.add_stream_to_page(
                dest_page,
                wr.add_object(
                    generic.StreamObject(stream_data=b"q"),
                ),
                prepend=True,
            )
            wr.add_stream_to_page(
                dest_page,
                wr.add_object(
                    generic.StreamObject(stream_data=b"Q"),
                ),
            )
        page_ref = wr.add_stream_to_page(
            dest_page, wr.add_object(stamp_wrapper_stream), resources
        )
        dims = (self.box.width, self.box.height)
        return page_ref, dims

    def as_appearances(self) -> AnnotAppearances:
        """
        Turn this stamp into an appearance dictionary for an annotation
        (or a form field widget), after rendering it.
        Only the normal appearance will be defined.

        :return:
            An instance of :class:`.AnnotAppearances`.
        """
        # TODO support defining overrides/extra's for the rollover/down
        #  appearances in some form
        stamp_ref = self.register()
        return AnnotAppearances(normal=stamp_ref)


class StaticContentStamp(BaseStamp):
    """Class representing stamps with static content."""

    def __init__(
        self,
        writer: BasePdfFileWriter,
        style: StaticStampStyle,
        box: layout.BoxConstraints,
    ):
        if not (box and box.height_defined and box.width_defined):
            raise layout.LayoutError(
                "StaticContentStamp requires a predetermined bounding box."
            )
        super().__init__(box=box, style=style, writer=writer)

    def _render_inner_content(self):
        return []


class TextStamp(BaseStamp):
    """
    Class that renders a text stamp as specified by an instance
    of :class:`.TextStampStyle`.
    """

    def __init__(
        self,
        writer: BasePdfFileWriter,
        style,
        text_params=None,
        box: Optional[layout.BoxConstraints] = None,
    ):
        super().__init__(box=box, style=style, writer=writer)
        self.text_params = text_params

        self.text_box: Optional[TextBox] = None

    def get_default_text_params(self):
        """
        Compute values for the default string interpolation parameters
        to be applied to the template string specified in the stamp
        style. This method does not take into account the ``text_params``
        init parameter yet.

        :return:
            A dictionary containing the parameters and their values.
        """
        ts = datetime.now(tz=tzlocal.get_localzone())
        return {
            'ts': ts.strftime(self.style.timestamp_format),
        }

    def _text_layout(self):
        # Set the contents of the text box
        self.text_box = tb = TextBox(
            self.style.text_box_style,
            writer=self.writer,
            resources=self.resources,
            box=None,
        )
        _text_params = self.get_default_text_params()
        if self.text_params is not None:
            _text_params.update(self.text_params)
        try:
            text = self.style.stamp_text % _text_params
        except KeyError as e:
            raise LayoutError(f"Stamp text parameter '{e.args[0]}' is missing")
        tb.content = text

        # Render the text box in its natural size, we'll deal with
        # the minutiae later
        return tb.render()

    def _inner_layout_natural_size(self):
        # render text
        text_commands = self._text_layout()

        inn_box = self.text_box.box
        return [text_commands], (inn_box.width, inn_box.height)

    def _inner_content_layout_rule(self):
        return self.style.inner_content_layout or DEFAULT_BOX_LAYOUT

    def _render_inner_content(self):
        command_stream = [b'q']

        # compute the inner bounding box
        inn_commands, (
            inn_width,
            inn_height,
        ) = self._inner_layout_natural_size()

        inner_layout = self._inner_content_layout_rule()

        bbox = self.box

        # position the inner box
        inn_position = inner_layout.fit(bbox, inn_width, inn_height)

        command_stream.append(inn_position.as_cm())
        command_stream.extend(inn_commands)
        command_stream.append(b'Q')

        return command_stream


class QRStamp(TextStamp):
    def __init__(
        self,
        writer: BasePdfFileWriter,
        url: str,
        style: QRStampStyle,
        text_params=None,
        box: Optional[layout.BoxConstraints] = None,
    ):
        super().__init__(writer, style, text_params=text_params, box=box)
        self.url = url
        self._qr_size = None

    def _inner_content_layout_rule(self):
        style = self.style
        if style.inner_content_layout is not None:
            return style.inner_content_layout

        # choose a reasonable default based on the QR code's relative position
        return style.qr_position.value

    def _inner_layout_natural_size(self) -> Tuple[List[bytes], Tuple[int, int]]:
        text_commands, (
            text_width,
            text_height,
        ) = super()._inner_layout_natural_size()

        qr_ref, natural_qr_size = self._qr_xobject()
        self.set_resource(
            category=content.ResourceType.XOBJECT,
            name=pdf_name('/QR'),
            value=qr_ref,
        )

        style = self.style
        stamp_box = self.box

        # To size the QR code, we proceed as follows:
        #  - If qr_inner_size is not None, use it
        #  - If the stamp has a fully defined bbox already,
        #    make sure it fits within the innseps, and it's not too much smaller
        #    than the text box
        #  - Else, scale down by DEFAULT_QR_SCALE and use that value
        #
        # Note: if qr_inner_size is defined AND the stamp bbox is available
        # already, scaling might still take effect depending on the inner layout
        # rule.
        innsep = style.innsep
        if style.qr_inner_size is not None:
            qr_size = style.qr_inner_size
        elif stamp_box.width_defined and stamp_box.height_defined:
            # ensure that the QR code doesn't shrink too much if the text
            # box is too tall.
            min_dim = min(
                max(stamp_box.height, text_height),
                max(stamp_box.width, text_width),
            )
            qr_size = min_dim - 2 * innsep
        else:
            qr_size = int(round(DEFAULT_QR_SCALE * natural_qr_size))

        qr_innunits_scale = qr_size / natural_qr_size
        qr_padded = qr_size + 2 * innsep
        # Next up: put the QR code and the text box together to get the
        # inner layout bounding box
        if style.qr_position.horizontal_flow:
            inn_width = qr_padded + text_width
            inn_height = max(qr_padded, text_height)
        else:
            inn_width = max(qr_padded, text_width)
            inn_height = qr_padded + text_height
        # grab the base layout rule from the QR position setting
        default_layout: layout.SimpleBoxLayoutRule = style.qr_position.value

        # Fill in the margins
        qr_layout_rule = layout.SimpleBoxLayoutRule(
            x_align=default_layout.x_align,
            y_align=default_layout.y_align,
            margins=layout.Margins.uniform(innsep),
            # There's no point in scaling here, the inner content canvas
            # is always big enough
            inner_content_scaling=layout.InnerScaling.NO_SCALING,
        )

        inner_box = layout.BoxConstraints(inn_width, inn_height)
        qr_inn_pos = qr_layout_rule.fit(inner_box, qr_size, qr_size)

        # we still need to take the axis reversal into account
        # (which also implies an adjustment in the y displacement)
        draw_qr_command = b'q %g 0 0 %g %g %g cm /QR Do Q' % (
            qr_inn_pos.x_scale * qr_innunits_scale,
            qr_inn_pos.y_scale * qr_innunits_scale,
            qr_inn_pos.x_pos,
            qr_inn_pos.y_pos,
        )

        # Time to put in the text box now
        if style.qr_position == QRPosition.LEFT_OF_TEXT:
            tb_margins = layout.Margins(
                left=qr_padded, right=0, top=0, bottom=0
            )
        elif style.qr_position == QRPosition.RIGHT_OF_TEXT:
            tb_margins = layout.Margins(
                right=qr_padded, left=0, top=0, bottom=0
            )
        elif style.qr_position == QRPosition.BELOW_TEXT:
            tb_margins = layout.Margins(
                bottom=qr_padded, right=0, left=0, top=0
            )
        else:
            tb_margins = layout.Margins(
                top=qr_padded, right=0, left=0, bottom=0
            )

        tb_layout_rule = layout.SimpleBoxLayoutRule(
            # flip around the alignment conventions of the default layout
            # to position the text box on the other side
            x_align=default_layout.x_align.flipped,
            y_align=default_layout.y_align.flipped,
            margins=tb_margins,
            inner_content_scaling=layout.InnerScaling.NO_SCALING,
        )

        # position the text box
        text_inn_pos = tb_layout_rule.fit(inner_box, text_width, text_height)

        commands = [draw_qr_command, b'q', text_inn_pos.as_cm()]
        commands.extend(text_commands)
        commands.append(b'Q')
        return commands, (inn_width, inn_height)

    def _qr_xobject(self):
        is_fancy = self.style.qr_inner_content is not None
        err_corr = (
            qrcode.constants.ERROR_CORRECT_H
            if is_fancy
            else qrcode.constants.ERROR_CORRECT_M
        )
        qr = qrcode.QRCode(error_correction=err_corr)
        qr.add_data(self.url)
        qr.make()

        if is_fancy:
            img = qr.make_image(
                image_factory=PdfFancyQRImage,
                version=qr.version,
                center_image=self.style.qr_inner_content,
            )
        else:
            img = qr.make_image(image_factory=PdfStreamQRImage)
        command_stream = img.render_command_stream()

        bbox_size = (qr.modules_count + 2 * qr.border) * qr.box_size
        qr_xobj = init_xobject_dictionary(command_stream, bbox_size, bbox_size)
        qr_xobj.compress()
        return self.writer.add_object(qr_xobj), bbox_size

    def get_default_text_params(self):
        tp = super().get_default_text_params()
        tp['url'] = self.url
        return tp

    def apply(
        self,
        dest_page,
        x,
        y,
        *,
        coords: CoordinateSystem = CoordinateSystem.PAGE_DEFAULT,
    ):
        page_ref, (w, h) = super().apply(dest_page, x, y, coords=coords)
        link_rect = (x, y, x + w, y + h)
        link_annot = generic.DictionaryObject(
            {
                pdf_name('/Type'): pdf_name('/Annot'),
                pdf_name('/Subtype'): pdf_name('/Link'),
                pdf_name('/Rect'): generic.ArrayObject(
                    list(map(generic.FloatObject, link_rect))
                ),
                pdf_name('/A'): generic.DictionaryObject(
                    {
                        pdf_name('/S'): pdf_name('/URI'),
                        pdf_name('/URI'): pdf_string(self.url),
                    }
                ),
            }
        )
        wr = self.writer
        assert wr is not None
        wr.register_annotation(page_ref, wr.add_object(link_annot))
        return page_ref, (w, h)


def _stamp_file(
    input_name: str,
    output_name: str,
    style: TextStampStyle,
    stamp_class,
    dest_page: int,
    x: int,
    y: int,
    **stamp_kwargs,
):
    with open(input_name, 'rb') as fin:
        pdf_out = IncrementalPdfFileWriter(fin, strict=False)
        stamp = stamp_class(writer=pdf_out, style=style, **stamp_kwargs)
        stamp.apply(dest_page, x, y)

        with open(output_name, 'wb') as out:
            pdf_out.write(out)


def text_stamp_file(
    input_name: str,
    output_name: str,
    style: TextStampStyle,
    dest_page: int,
    x: int,
    y: int,
    text_params=None,
):
    """
    Add a text stamp to a file.

    :param input_name:
        Path to the input file.
    :param output_name:
        Path to the output file.
    :param style:
        Text stamp style to use.
    :param dest_page:
        Index of the page to which the stamp is to be applied (starting at `0`).
    :param x:
        Horizontal position of the stamp's lower left corner on the page.
    :param y:
        Vertical position of the stamp's lower left corner on the page.
    :param text_params:
        Additional parameters for text template interpolation.
    """
    _stamp_file(
        input_name,
        output_name,
        style,
        TextStamp,
        dest_page,
        x,
        y,
        text_params=text_params,
    )


def qr_stamp_file(
    input_name: str,
    output_name: str,
    style: QRStampStyle,
    dest_page: int,
    x: int,
    y: int,
    url: str,
    text_params=None,
):
    """
    Add a QR stamp to a file.

    :param input_name:
        Path to the input file.
    :param output_name:
        Path to the output file.
    :param style:
        QR stamp style to use.
    :param dest_page:
        Index of the page to which the stamp is to be applied (starting at `0`).
    :param x:
        Horizontal position of the stamp's lower left corner on the page.
    :param y:
        Vertical position of the stamp's lower left corner on the page.
    :param url:
        URL for the QR code to point to.
    :param text_params:
        Additional parameters for text template interpolation.
    """

    _stamp_file(
        input_name,
        output_name,
        style,
        QRStamp,
        dest_page,
        x,
        y,
        url=url,
        text_params=text_params,
    )


STAMP_ART_CONTENT = content.RawContent(
    box=layout.BoxConstraints(width=100, height=100),
    data=b'''
q 1 0 0 -1 0 100 cm 
0.603922 0.345098 0.54902 rg
3.699 65.215 m 3.699 65.215 2.375 57.277 7.668 51.984 c 12.957 46.695 27.512
 49.34 39.418 41.402 c 39.418 41.402 31.48 40.078 32.801 33.465 c 34.125
 26.852 39.418 28.172 39.418 24.203 c 39.418 20.234 30.156 17.59 30.156
14.945 c 30.156 12.297 28.465 1.715 50 1.715 c 71.535 1.715 69.844 12.297
 69.844 14.945 c 69.844 17.59 60.582 20.234 60.582 24.203 c 60.582 28.172
 65.875 26.852 67.199 33.465 c 68.52 40.078 60.582 41.402 60.582 41.402
c 72.488 49.34 87.043 46.695 92.332 51.984 c 97.625 57.277 96.301 65.215
 96.301 65.215 c h f
3.801 68.734 92.398 7.391 re f
3.801 79.512 92.398 7.391 re f
3.801 90.289 92.398 7.391 re f
Q
''',
)
"""
Hardcoded stamp background that will render a stylised image of a stamp using 
PDF graphics operators (see below).

.. image:: images/stamp-background.svg
   :alt: Standard stamp background
   :align: center
   
"""



"""
Utility for writing incremental updates to existing PDF files.
"""

import os
from typing import Optional, Union

from . import generic, misc
from .crypt import EnvelopeKeyDecrypter
from .generic import pdf_name
from .metadata.model import DocumentMetadata
from .reader import PdfFileReader, parse_catalog_version
from .writer import BasePdfFileWriter

__all__ = ['IncrementalPdfFileWriter']


class IncrementalPdfFileWriter(BasePdfFileWriter):
    """Class to incrementally update existing files.

    This :class:`~.writer.BasePdfFileWriter` subclass encapsulates a
    :class:`~.reader.PdfFileReader` instance in addition to exposing an
    interface to add and modify PDF objects.

    Incremental updates to a PDF file append modifications to the end of the
    file. This is critical when the original file contents are not to be
    modified directly (e.g. when it contains digital signatures).
    It has the additional advantage of providing an automatic audit trail of
    sorts.

    :param input_stream:
        Input stream to read current revision from.
    :param strict:
        Ingest the source file in strict mode. The default is ``True``.
    :param prev:
        Explicitly pass in a PDF reader. This parameter is internal API.
    """

    IO_CHUNK_SIZE = 4096
    _force_write_when_empty = False

    def __init__(
        self, input_stream, prev: Optional[PdfFileReader] = None, strict=True
    ):
        self.input_stream = input_stream
        if prev is None:
            prev = PdfFileReader(input_stream, strict=strict)
        self.prev = prev
        self.trailer = trailer = prev.trailer
        root_ref = trailer.raw_get('/Root')
        try:
            info_ref = trailer.raw_get('/Info')
        except KeyError:
            # rare, but it can happen. /Info is not a required entry
            info_ref = None
        document_id = self.__class__._handle_id(prev)
        super().__init__(
            root_ref,
            info_ref,
            document_id,
            obj_id_start=trailer['/Size'],
            stream_xrefs=prev.has_xref_stream,
        )
        if self._info is not None:
            self.trailer['/Info'] = self._info
        self._resolves_objs_from = (self, prev)
        input_ver = self.prev.input_version
        if input_ver > self.output_version:
            self.output_version = input_ver
        self.ensure_output_version(self.__class__.output_version)

        self.security_handler = prev.security_handler
        if self.security_handler is not None:
            self._encrypt = prev.trailer.raw_get("/Encrypt")

    @classmethod
    def from_reader(cls, reader: PdfFileReader) -> 'IncrementalPdfFileWriter':
        """
        Instantiate an incremental writer from a PDF file reader.

        :param reader:
            A :class:`.PdfFileReader` object with a PDF to extend.
        """

        return cls(reader.stream, prev=reader)

    def ensure_output_version(self, version):
        # check header
        if self.prev.input_version >= version:
            return
        # check root
        root = self.root
        try:
            ver = root[pdf_name('/Version')]
            cur_version = parse_catalog_version(ver)
            if cur_version is not None and cur_version >= version:
                return
        except (KeyError, ValueError, TypeError):
            pass
        version_str = pdf_name('/%d.%d' % version)
        root[pdf_name('/Version')] = version_str
        self.update_root()
        self.output_version = version

    @classmethod
    def _handle_id(cls, prev):
        # There are a number of issues at play here
        #  - Documents *should* have a unique id, but it's not a strict
        #    requirement unless the document is encrypted.
        #  - We are updating an existing document, but the result is not the
        #    same document. Hence, we want to assign an ID to this document that
        #    is not the same as the one on the existing document.
        #  - The first part of the ID is part of the key derivation used to
        #    to encrypt documents. Since we need to encrypt the file using
        #    the same cryptographic data as the original, we cannot change
        #    this value if it is present (cf. § 7.6.3.3 in ISO 32000).
        #    Even when no encryption is involved, changing this part violates
        #    the spec (cf. § 14.4 in loc. cit.)

        # noinspection PyArgumentList
        id2 = generic.ByteStringObject(os.urandom(16))
        try:
            id1, _ = prev.trailer["/ID"]
            # is this a bug in PyPDF2?
            if isinstance(id1, generic.TextStringObject):
                # noinspection PyArgumentList
                id1 = generic.ByteStringObject(id1.original_bytes)
        except KeyError:
            # no primary ID present, so generate one
            # noinspection PyArgumentList
            id1 = generic.ByteStringObject(os.urandom(16))
        return generic.ArrayObject([id1, id2])

    def get_object(self, ido, as_metadata_stream: bool = False):
        try:
            return super().get_object(
                ido, as_metadata_stream=as_metadata_stream
            )
        except KeyError:
            return self.prev.get_object(
                ido, as_metadata_stream=as_metadata_stream
            )

    def mark_update(
        self, obj_ref: Union[generic.Reference, generic.IndirectObject]
    ):
        ix = (obj_ref.generation, obj_ref.idnum)
        self.objects[ix] = obj_ref.get_object()

    # TODO: this new API allows me to simplify a lot of bookkeeping
    #  in the library
    def update_container(self, obj: generic.PdfObject):
        container_ref = obj.container_ref
        if container_ref is None:
            # this means that in all likelihood, the object was added by this
            # writer, and is therefore about to be written anyway.
            return
        if isinstance(container_ref, generic.TrailerReference):
            # nothing to do, the trailer is always written
            return
        elif isinstance(container_ref, generic.Reference):
            self.mark_update(container_ref)
            return
        raise TypeError  # pragma: nocover

    def update_root(self):
        self.mark_update(self._root)

    def _write_header(self, stream):
        # copy the original data to the output
        input_pos = self.input_stream.tell()
        self.input_stream.seek(0)
        misc.chunked_write(
            bytearray(self.IO_CHUNK_SIZE), self.input_stream, stream
        )
        self.input_stream.seek(input_pos)

    def set_info(
        self,
        info: Optional[Union[generic.IndirectObject, generic.DictionaryObject]],
    ):
        info = super().set_info(info)
        if info is not None:
            # also update our trailer
            self.trailer['/Info'] = info
        else:
            del self.trailer['/Info']
        return info

    def _populate_trailer(self, trailer):
        trailer.update(self.trailer.flatten())
        super()._populate_trailer(trailer)
        trailer[pdf_name('/Prev')] = generic.NumberObject(
            self.prev.last_startxref
        )
        if self.prev.encrypted and not self.security_handler.is_authenticated():
            # removing encryption in an incremental update is impossible
            raise misc.PdfWriteError(
                'Cannot update this document without encryption credentials '
                'from the original. Please call encrypt() with the password '
                'of the original file before calling write().'
            )

    def set_custom_trailer_entry(
        self, key: generic.NameObject, value: generic.PdfObject
    ):
        """
        Set a custom, unmanaged entry in the document trailer or cross-reference
        stream dictionary.

        .. warning::
            Calling this method to set an entry that is managed by pyHanko
            internally (info dictionary, document catalog, etc.) has undefined
            results.

        :param key:
            Dictionary key to use in the trailer.
        :param value:
            Value to set
        """
        self.trailer[key] = value

    def write(self, stream):
        if not self.objects and not self._force_write_when_empty:
            # just write the original and then bail
            self._write_header(stream)
            return
        self._prep_dom_for_writing()
        super().write(stream)

    @property
    def document_meta_view(self) -> DocumentMetadata:
        return self._meta.view_over(self.prev.document_meta_view)

    def _write_updated_section(self, stream):
        """
        Only write the updated and new objects to the designated output stream.

        The new PDF file can then be put together by concatenating the original
        input with the generated output.

        .. danger::
            The offsets in this output will typically be wrong unless the missing
            previous section is somehow taken into account.

        .. danger::
            Object graph finalisation will not run.

        :param stream:
            Output stream to write to.
        """
        self._write(stream, skip_header=True)

    def write_in_place(self):
        """
        Write the updated file contents in-place to the same stream as
        the input stream.
        This obviously requires a stream supporting both reading and writing
        operations.
        """
        self._prep_dom_for_writing()
        if self._need_mac_on_write:
            from pyhanko.pdf_utils.crypt import pdfmac

            pdfmac.add_standalone_mac(self, in_place=True)
        else:
            stream = self.prev.stream
            stream.seek(0, os.SEEK_END)
            self._write_updated_section(stream)

    def encrypt(self, user_pwd):
        """Method to handle updates to encrypted files.

        This method handles decrypting of the original file, and makes sure
        the resulting updated file is encrypted in a compatible way.
        The standard mandates that updates to encrypted files be effected using
        the same encryption settings. In particular, incremental updates
        cannot remove file encryption.

        :param user_pwd:
            The original file's user password.

        :raises PdfReadError:
            Raised when there is a problem decrypting the file.
        """

        prev = self.prev
        result = prev.decrypt(user_pwd)

        # take care to use the same encryption algorithm as the underlying file
        self._encrypt = prev.trailer.raw_get("/Encrypt")
        return result

    def encrypt_pubkey(self, credential: EnvelopeKeyDecrypter):
        """Method to handle updates to files encrypted using public-key
        encryption.

        The same caveats as :meth:`encrypt` apply here.


        :param credential:
            The :class:`.EnvelopeKeyDecrypter` handling the recipient's
            private key.

        :raises PdfReadError:
            Raised when there is a problem decrypting the file.
        """

        prev = self.prev
        result = prev.decrypt_pubkey(credential)
        self._encrypt = prev.trailer.raw_get("/Encrypt")
        return result



import os
from io import BytesIO

import pytest
from PIL import Image
from pyhanko.pdf_utils import generic, images
from pyhanko.pdf_utils.generic import pdf_name
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter

from .samples import *

IMG_DIR = f'{TEST_DIR}/data/img'


@pytest.mark.parametrize('infile', ['stamp.png', 'stamp-indexed.png'])
def test_image_direct_embed(infile):
    img = Image.open(os.path.join(IMG_DIR, infile))

    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    image_ref = images.pil_image(img, w)
    page_ref, resources = w.find_page_for_modification(0)
    resources[pdf_name('/XObject')] = generic.DictionaryObject(
        {pdf_name('/Img'): image_ref}
    )
    w.update_container(resources)
    content_stream: generic.StreamObject = page_ref.get_object()['/Contents']
    content_stream._data = (
        content_stream.data + b' q 50 0 0 50 5 5 cm /Img Do Q'
    )
    content_stream._encoded_data = None
    w.update_container(content_stream)
    w.write_in_place()

    # TODO flatten and do a visual comparison

    image_obj = image_ref.get_object()
    if 'indexed' in infile:
        assert '/SMask' not in image_obj
    else:
        assert '/SMask' in image_obj


@pytest.mark.parametrize('infile', ['stamp.png', 'stamp-indexed.png'])
def test_image_content_embed(infile):
    path = os.path.join(IMG_DIR, infile)

    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    img_content = images.PdfImage(
        image=path,
        writer=w,
        opacity=0.6,
    )
    img_content.add_to_page(writer=w, page_ix=0, prepend=True)
    w.write_in_place()



from .xmldsig_core import (
    CanonicalizationMethod,
    CanonicalizationMethodType,
    DigestMethod,
    DigestMethodType,
    DigestValue,
    DSAKeyValue,
    DSAKeyValueType,
    KeyInfo,
    KeyInfoType,
    KeyName,
    KeyValue,
    KeyValueType,
    Manifest,
    ManifestType,
    MgmtData,
    Object,
    ObjectType,
    PGPData,
    PGPDataType,
    Reference,
    ReferenceType,
    RetrievalMethod,
    RetrievalMethodType,
    RSAKeyValue,
    RSAKeyValueType,
    Signature,
    SignatureMethod,
    SignatureMethodType,
    SignatureProperties,
    SignaturePropertiesType,
    SignatureProperty,
    SignaturePropertyType,
    SignatureType,
    SignatureValue,
    SignatureValueType,
    SignedInfo,
    SignedInfoType,
    SPKIData,
    SPKIDataType,
    Transform,
    Transforms,
    TransformsType,
    TransformType,
    X509Data,
    X509DataType,
    X509IssuerSerialType,
)

__all__ = [
    "CanonicalizationMethod",
    "CanonicalizationMethodType",
    "DSAKeyValue",
    "DSAKeyValueType",
    "DigestMethod",
    "DigestMethodType",
    "DigestValue",
    "KeyInfo",
    "KeyInfoType",
    "KeyName",
    "KeyValue",
    "KeyValueType",
    "Manifest",
    "ManifestType",
    "MgmtData",
    "Object",
    "ObjectType",
    "PGPData",
    "PGPDataType",
    "RSAKeyValue",
    "RSAKeyValueType",
    "Reference",
    "ReferenceType",
    "RetrievalMethod",
    "RetrievalMethodType",
    "SPKIData",
    "SPKIDataType",
    "Signature",
    "SignatureMethod",
    "SignatureMethodType",
    "SignatureProperties",
    "SignaturePropertiesType",
    "SignatureProperty",
    "SignaturePropertyType",
    "SignatureType",
    "SignatureValue",
    "SignatureValueType",
    "SignedInfo",
    "SignedInfoType",
    "Transform",
    "TransformType",
    "Transforms",
    "TransformsType",
    "X509Data",
    "X509DataType",
    "X509IssuerSerialType",
]
